#! /usr/bin/env python

"""
Module with the function of merit definitions for the NEGFC optimization.
"""

__author__ = 'O. Wertz, Carlos Alberto Gomez Gonzalez, Valentin Christiaens'
__all__ = []

import numpy as np
from hciplot import plot_frames
import photutils
from skimage.draw import disk
from ..metrics import cube_inject_companions, snr
from ..var import (frame_center, get_annular_wedge, cube_filter_highpass, dist,
                   get_circle)
from ..pca import pca_annulus, pca_annular, pca
from ..preproc import cube_crop_frames
from ..conf import check_array

def chisquare(modelParameters, cube, angs, plsc, psfs_norm, fwhm, annulus_width,  
              aperture_radius, initialState, ncomp, cube_ref=None, 
              svd_mode='lapack', scaling=None, fmerit='sum', collapse='median',
              algo=pca_annulus, delta_rot=1, imlib='opencv', 
              interpolation='lanczos4', algo_options={}, transmission=None, 
              mu_sigma=None, weights=None, force_rPA=False, debug=False):
    """
    Calculate the reduced chi2:
    \chi^2_r = \frac{1}{N-3}\sum_{j=1}^{N} |I_j|,
    where N is the number of pixels within a circular aperture centered on the 
    first estimate of the planet position, and I_j the j-th pixel intensity.
    
    Parameters
    ----------    
    modelParameters: tuple
        The model parameters, typically (r, theta, flux).
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    plsc: float
        The platescale, in arcsec per pixel.
    psfs_norm: numpy.array
        The scaled psf expressed as a numpy.array.    
    fwhm : float
        The FHWM in pixels.
    annulus_width: int, optional
        The width in pixels of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    initialState: numpy.array
        Position (r, theta) of the circular aperture center.
    ncomp: int or None
        The number of principal components for PCA-based algorithms.
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.         
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    algo_options: dict, opt
        Dictionary with additional parameters related to the algorithm 
        (e.g. tol, min_frames_lib, max_frames_lib). If 'algo' is not a vip
        routine, this dict should contain all necessary arguments apart from
        the cube and derotation angles. Note: arguments such as ncomp, svd_mode,
        scaling, imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for consistency
        with older versions of vip). 
    transmission: numpy array, optional
        Array with 2 columns. First column is the radial separation in pixels. 
        Second column is the off-axis transmission (between 0 and 1) at the 
        radial separation given in column 1.
    mu_sigma: tuple of 2 floats or None, opt
        If set to None: not used, and falls back to original version of the 
        algorithm, using fmerit.
        If set to anything but None: will compute the mean and standard 
        deviation of pixel intensities in an annulus centered on the location 
        of the companion, excluding the area directly adjacent to the companion.
    weights : 1d array, optional
        If provided, the negative fake companion fluxes will be scaled according
        to these weights before injection in the cube. Can reflect changes in 
        the observing conditions throughout the sequence.
    force_rPA: bool, optional
        Whether to only search for optimal flux, provided (r,PA).
    debug: bool, opt
        Whether to debug and plot the post-processed frame after injection of 
        the negative fake companion.
        
    Returns
    -------
    out: float
        The reduced chi squared.
        
    """
    if force_rPA:
        r, theta = initialState
        flux_tmp = modelParameters[0]
    else:
        try:
            r, theta, flux_tmp = modelParameters
        except TypeError:
            msg = 'modelParameters must be a tuple, {} was given'
            print(msg.format(type(modelParameters)))

    ## set imlib for rotation and shift
    if imlib == 'opencv':
        imlib_sh = imlib
        imlib_rot = imlib
    elif imlib == 'skimage' or imlib == 'ndimage-interp':
        imlib_sh = 'ndimage-interp'
        imlib_rot = 'skimage'
    elif imlib == 'vip-fft' or imlib == 'ndimage-fourier':
        imlib_sh = 'ndimage-fourier'
        imlib_rot = 'vip-fft'
    else:
        raise TypeError("Interpolation not recognized.")

    if weights is None:   
        flux = -flux_tmp
        norm_weights=weights
    else:
        flux = -flux_tmp*weights
        norm_weights = weights/np.sum(weights)

    # Create the cube with the negative fake companion injected
    cube_negfc = cube_inject_companions(cube, psfs_norm, angs, flevel=flux,
                                        plsc=plsc, rad_dists=[r], n_branches=1,
                                        theta=theta, imlib=imlib_sh, 
                                        verbose=False, 
                                        interpolation=interpolation, 
                                        transmission=transmission)
                                      
    # Perform PCA and extract the zone of interest
    res = get_values_optimize(cube_negfc, angs, ncomp, annulus_width,
                              aperture_radius, fwhm, initialState[0],
                              initialState[1], cube_ref=cube_ref,
                              svd_mode=svd_mode, scaling=scaling, algo=algo,
                              delta_rot=delta_rot, collapse=collapse, 
                              algo_options=algo_options, weights=norm_weights, 
                              imlib=imlib_rot, interpolation=interpolation, 
                              debug=debug)
    
    if debug and collapse is not None:
        values, frpca = res
        plot_frames(frpca)
    else:
        values = res
    
    # Function of merit
    if mu_sigma is None:
        # old version - delete?
        if fmerit == 'sum':
            chi = np.sum(np.abs(values))
        elif fmerit == 'stddev':
            values = values[values != 0]
            chi = np.std(values)*values.size # TODO: test std**2
        else:
            raise RuntimeError('fmerit choice not recognized.')
    else:
        # true expression of a gaussian log probability
        mu = mu_sigma[0]
        sigma = mu_sigma[1]
        chi = np.sum(np.power(mu-values,2)/sigma**2)   
        
    return chi


def get_values_optimize(cube, angs, ncomp, annulus_width, aperture_radius, 
                        fwhm, r_guess, theta_guess, cube_ref=None, 
                        svd_mode='lapack', scaling=None, algo=pca_annulus, 
                        delta_rot=1, imlib='opencv', interpolation='lanczos4',
                        collapse='median', algo_options={}, weights=None, 
                        debug=False):
    """ Extracts a PCA-ed annulus from the cube and returns the flux values of
    the pixels included in a circular aperture centered at a given position.
    
    Parameters
    ----------
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.
    ncomp: int or None
        The number of principal components for PCA-based algorithms.
    annulus_width: float
        The width in pixels of the annulus on which the PCA is performed.
    aperture_radius: float
        The radius in fwhm of the circular aperture.
    fwhm: float
        Value of the FWHM of the PSF.
    r_guess: float
        The radial position of the center of the circular aperture. This 
        parameter is NOT the radial position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.
    theta_guess: float
        The angular position of the center of the circular aperture. This 
        parameter is NOT the angular position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.  
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {None, 'temp-mean', 'temp-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done.
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is returned.
    algo_options: dict, opt
        Dictionary with additional parameters related to the algorithm 
        (e.g. tol, min_frames_lib, max_frames_lib). If 'algo' is not a vip
        routine, this dict should contain all necessary arguments apart from
        the cube and derotation angles. Note: arguments such as ncomp, svd_mode,
        scaling, imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for consistency
        with older versions of vip). 
    weights : 1d array, optional
        If provided, the negative fake companion fluxes will be scaled according
        to these weights before injection in the cube. Can reflect changes in 
        the observing conditions throughout the sequence.
    debug: boolean
        If True, the cube is returned along with the values.        
        
    Returns
    -------
    values: numpy.array
        The pixel values in the circular aperture after the PCA process.

    If debug is True and collapse non-None, the PCA frame is also returned.
        
    """
    centy_fr, centx_fr = frame_center(cube[0])
    posy = r_guess * np.sin(np.deg2rad(theta_guess)) + centy_fr
    posx = r_guess * np.cos(np.deg2rad(theta_guess)) + centx_fr
    halfw = max(aperture_radius*fwhm, annulus_width/2)

    # Checking annulus/aperture sizes. Assuming square frames
    msg = 'The annulus and/or the circular aperture used by the NegFC falls '
    msg += 'outside the FOV. Try increasing the size of your frames or '
    msg += 'decreasing the annulus or aperture size.'
    msg += 'rguess: {:.0f}px; centx_fr: {:.0f}px'.format(r_guess,centx_fr)
    msg += 'halfw: {:.0f}px'.format(halfw)
    if r_guess > centx_fr-halfw:# or r_guess <= halfw:
        raise RuntimeError(msg)

    ncomp = algo_options.get('ncomp',ncomp)
    svd_mode = algo_options.get('svd_mode',svd_mode)
    scaling = algo_options.get('scaling',scaling)
    imlib = algo_options.get('imlib',imlib)
    interpolation = algo_options.get('interpolation',interpolation)
    collapse = algo_options.get('collapse',collapse)


    if algo == pca_annulus:
        pca_res = pca_annulus(cube, angs, ncomp, annulus_width, r_guess, 
                              cube_ref, svd_mode, scaling, imlib=imlib,
                              interpolation=interpolation, collapse=collapse,
                              weights=weights)
        
    elif algo == pca_annular:
                
        tol = algo_options.get('tol',1e-1)
        min_frames_lib=algo_options.get('min_frames_lib',2)
        max_frames_lib=algo_options.get('max_frames_lib',200)
        radius_int = max(1,int(np.floor(r_guess-annulus_width/2)))
        radius_int = algo_options.get('radius_int', radius_int)
        # crop cube to just be larger than annulus => FASTER PCA
        crop_sz = int(2*np.ceil(radius_int+annulus_width+1))
        if not crop_sz %2:
            crop_sz+=1
        if crop_sz < cube.shape[1] and crop_sz < cube.shape[2]:
            pad = int((cube.shape[1]-crop_sz)/2)
            crop_cube = cube_crop_frames(cube, crop_sz, verbose=False)
        else:
            crop_cube = cube


        pca_res_tmp = pca_annular(crop_cube, angs, radius_int=radius_int, 
                                  fwhm=fwhm, asize=annulus_width, 
                                  delta_rot=delta_rot, ncomp=ncomp, 
                                  svd_mode=svd_mode, scaling=scaling, 
                                  imlib=imlib, interpolation=interpolation,
                                  collapse=collapse, weights=weights, tol=tol, 
                                  min_frames_lib=min_frames_lib, 
                                  max_frames_lib=max_frames_lib, 
                                  full_output=False, verbose=False)
        # pad again now                      
        pca_res = np.pad(pca_res_tmp,pad,mode='constant',constant_values=0)
        
    elif algo == pca: 
        scale_list = algo_options.get('scale_list',None)
        ifs_collapse_range = algo_options.get('ifs_collapse_range','all')
        nproc = algo_options.get('nproc','all')
        pca_res = pca(cube, angs, cube_ref, scale_list, ncomp, 
                      svd_mode=svd_mode, scaling=scaling, imlib=imlib,
                      interpolation=interpolation, collapse=collapse,
                      ifs_collapse_range=ifs_collapse_range, nproc=nproc,
                      weights=weights, verbose=False)
    else:
        algo_args = algo_options
        pca_res = algo(cube, angs, **algo_args)
                                  
    indices = disk((posy, posx), radius=aperture_radius*fwhm)
    yy, xx = indices

    if collapse is None:
        values = pca_res[:, yy, xx].ravel()
    else:
        values = pca_res[yy, xx].ravel()

    if debug and collapse is not None:
        return values, pca_res
    else:
        return values


def get_mu_and_sigma(cube, angs, ncomp, annulus_width, aperture_radius, 
                     fwhm, r_guess, theta_guess, cube_ref=None, wedge=None,
                     svd_mode='lapack', scaling=None, algo=pca_annulus, 
                     delta_rot=1, imlib='opencv', interpolation='lanczos4',
                     collapse='median', weights=None, as_snr=False, 
                     algo_options={}):
    """ Extracts the mean and standard deviation of pixel intensities in an
    annulus of the PCA-ADI image obtained with 'algo', in the part of a defined 
    wedge that is not overlapping with PA_pl+-delta_PA.
    
    Parameters
    ----------
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.
    ncomp: int or None
        The number of principal components for PCA-based algorithms.
    annulus_width: float
        The width in pixels of the annulus on which the PCA is performed.
    aperture_radius: float
        The radius in fwhm of the circular aperture.
    fwhm: float
        Value of the FWHM of the PSF.
    r_guess: float
        The radial position of the center of the circular aperture. This 
        parameter is NOT the radial position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.
    theta_guess: float
        The angular position of the center of the circular aperture. This 
        parameter is NOT the angular position of the candidate associated to the 
        Markov chain, but should be the fixed initial guess.  
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    wedge: tuple, opt
        Range in theta where the mean and standard deviation are computed in an 
        annulus defined in the PCA image. If None, it will be calculated 
        automatically based on initial guess and derotation angles to avoid.
        If some disc signal is present elsewhere in the annulus, it is 
        recommended to provide wedge manually. The provided range should be 
        continuous and >0. E.g. provide (270, 370) to consider a PA range 
        between [-90,+10].
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {None, 'temp-mean', 'temp-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done.
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is returned.
    weights: 1d numpy array or list, optional
        Weights to be applied for a weighted mean. Need to be provided if 
        collapse mode is 'wmean'.
    algo_options: dict, opt
        Dictionary with additional parameters related to the algorithm 
        (e.g. tol, min_frames_lib, max_frames_lib). If 'algo' is not a vip
        routine, this dict should contain all necessary arguments apart from
        the cube and derotation angles. Note: arguments such as ncomp, svd_mode,
        scaling, imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for consistency
        with older versions of vip). 
        
    Returns
    -------
    values: numpy.array
        The pixel values in the circular aperture after the PCA process.
        
    """
    centy_fr, centx_fr = frame_center(cube[0])
    halfw = max(aperture_radius*fwhm, annulus_width/2)

    # Checking annulus/aperture sizes. Assuming square frames
    msg = 'The annulus and/or the circular aperture used by the NegFC falls '
    msg += 'outside the FOV. Try increasing the size of your frames or '
    msg += 'decreasing the annulus or aperture size.'
    msg += 'rguess: {:.0f}px; centx_fr: {:.0f}px'.format(r_guess,centx_fr)
    msg += 'halfw: {:.0f}px'.format(halfw)
    if r_guess > centx_fr-halfw:# or r_guess <= halfw:
        raise RuntimeError(msg)

    ncomp = algo_options.get('ncomp',ncomp)                       
    svd_mode = algo_options.get('svd_mode',svd_mode)
    scaling = algo_options.get('scaling',scaling)
    imlib = algo_options.get('imlib',imlib)
    interpolation = algo_options.get('interpolation',interpolation)
    collapse = algo_options.get('collapse',collapse)

    radius_int = max(1,int(np.floor(r_guess-annulus_width/2)))
    radius_int = algo_options.get('radius_int', radius_int)


    # not recommended, except if large-scale residual sky present (NIRC2-L')
    hp_filter = algo_options.get('hp_filter',None)
    hp_kernel = algo_options.get('hp_kernel',None)
    if hp_filter is not None:
        if 'median' in hp_filter:
            cube = cube_filter_highpass(cube, mode=hp_filter, 
                                        median_size=hp_kernel)
        elif "gauss" in hp_filter:
            cube = cube_filter_highpass(cube, mode=hp_filter, 
                                        fwhm_size=hp_kernel)
        else:
            cube = cube_filter_highpass(cube, mode=hp_filter, 
                                        kernel_size=hp_kernel)
    
    if algo == pca_annulus:
        pca_res = pca_annulus(cube, angs, ncomp, annulus_width, r_guess, 
                              cube_ref, svd_mode, scaling, imlib=imlib,
                              interpolation=interpolation, collapse=collapse, 
                              weights=weights)
        
    elif algo == pca_annular:                
        tol = algo_options.get('tol',1e-1)
        min_frames_lib=algo_options.get('min_frames_lib',2)
        max_frames_lib=algo_options.get('max_frames_lib',200)       
        # crop cube to just be larger than annulus => FASTER PCA
        crop_sz = int(2*np.ceil(radius_int+annulus_width+1))
        if not crop_sz %2:
            crop_sz+=1
        if crop_sz < cube.shape[1] and crop_sz < cube.shape[2]:
            pad = int((cube.shape[1]-crop_sz)/2)
            crop_cube = cube_crop_frames(cube, crop_sz, verbose=False)
        else:
            crop_cube = cube

        pca_res_tmp = pca_annular(crop_cube, angs, radius_int=radius_int, 
                                  fwhm=fwhm, asize=annulus_width, 
                                  delta_rot=delta_rot, ncomp=ncomp, 
                                  svd_mode=svd_mode, scaling=scaling, 
                                  imlib=imlib, interpolation=interpolation,
                                  collapse=collapse, tol=tol, 
                                  min_frames_lib=min_frames_lib, 
                                  max_frames_lib=max_frames_lib, 
                                  full_output=False, verbose=False, 
                                  weights=weights)
        # pad again now                      
        pca_res = np.pad(pca_res_tmp,pad,mode='constant',constant_values=0)
        
    elif algo == pca:
        scale_list = algo_options.get('scale_list',None)
        ifs_collapse_range = algo_options.get('ifs_collapse_range','all')
        nproc = algo_options.get('nproc','all')
        
        pca_res = pca(cube, angs, cube_ref, scale_list, ncomp, 
                      svd_mode=svd_mode, scaling=scaling, imlib=imlib,
                      interpolation=interpolation, collapse=collapse,
                      ifs_collapse_range=ifs_collapse_range, nproc=nproc,
                      weights=weights, verbose=False)
        
    else:
        algo_args = algo_options
        pca_res = algo(cube, angs, **algo_args)
        
    if wedge is None:
        delta_theta = np.amax(angs)-np.amin(angs)
        if delta_theta>150:
            delta_theta = 150 # if too much rotation, be less conservative
            
        theta_ini = (theta_guess+delta_theta)%360
        theta_fin = theta_ini+delta_theta
        wedge = (theta_ini,theta_fin)
    elif len(wedge)==2:
        if wedge[0]>wedge[1]:
            msg = '2nd value of wedge smaller than first one => 360 was added'
            print(msg)
            wedge = (wedge[0],wedge[1]+360)
    else:
        raise TypeError("Wedge should have exactly 2 values")
    
    if as_snr:
        source_xy = (centx_fr+r_guess*np.cos(np.deg2rad(theta_guess)), 
                     centy_fr+r_guess*np.sin(np.deg2rad(theta_guess)))
        print(source_xy)
        res_snr = snr(pca_res, source_xy, fwhm, full_output=True, 
                      exclude_negative_lobes=True)
        fluxes = res_snr[-2]
        fluxes = fluxes[1:-1]
        n2 = len(fluxes)
        mu = fluxes.mean()
        #sigma = fluxes.std()*np.sqrt(1+1/n2)
        sigma = get_sigma(pca_res, source_xy, fwhm, aperture_radius,
                          exclude_negative_lobes=True)
    else:
        indices = get_annular_wedge(pca_res, radius_int, annulus_width, 
                                    wedge=wedge)
        
        yy, xx = indices
        mu = np.mean(pca_res[yy, xx])
        # effective number of samples = Npx/Nfwhm instead of Npx
        # trick is to adapt ddof parameter in std function:
        npx = len(yy)
        area = np.pi*(fwhm/2)**2
        ddof = min(int(npx*(1.-(1./area)))+1, npx-1)
        sigma = np.std(pca_res[yy, xx], ddof=ddof)

    return mu, sigma


def get_sigma(array, source_xy, fwhm, aperture, full_output=False, 
              exclude_negative_lobes=False, verbose=False):
    """
    Estimate the average standard deviation in apertures similar as the one
    used for NEGFC minimization. This is done through the mean of as many 
    independent apertures of the same size at the same radius.
    
    Parameters
    ----------
    array : numpy ndarray, 2d
        Post-processed frame where we want to measure S/N.
    source_xy : tuple of floats
        X and Y coordinates of the planet or test speckle.
    fwhm : float
        Size in pixels of the FWHM.
    aperture: float
        Size of aperture in FWHM.
    full_output : bool, optional
        If True returns back the S/N value, the y, x input coordinates, noise
        and flux.
    exclude_negative_lobes : bool, opt
        Whether to include the adjacent aperture lobes to the tested location 
        or not. Can be set to True if the image shows significant neg lobes.
    verbose: bool, optional
        Chooses whether to print some output or not.

    Returns
    -------
    [if full_output=True:]
    sourcey : numpy ndarray
        [full_output=True] Input coordinates (``source_xy``) in Y.
    sourcex : numpy ndarray
        [full_output=True] Input coordinates (``source_xy``) in X.
    f_source : float
        [full_output=True] Flux in test elemnt.
    fluxes : numpy ndarray
        [full_output=True] Background apertures fluxes.
    [always:]    
    snr_vale : float
        Value of the S/N for the given test resolution element.
    """
    check_array(array, dim=2, msg='array')
    if not isinstance(source_xy, tuple):
        raise TypeError("`source_xy` must be a tuple of floats")

    sourcex, sourcey = source_xy
    centery, centerx = frame_center(array)
    sep = dist(centery, centerx, float(sourcey), float(sourcex))

    if not sep > (aperture*fwhm)+1:
        raise RuntimeError('`source_xy` is too close to the frame center')

    sens = 'clock'  # counterclock

    angle = np.arcsin(aperture*fwhm/sep)*2
    number_apertures = int(np.floor(2*np.pi/angle))
    yy = np.zeros((number_apertures))
    xx = np.zeros((number_apertures))
    cosangle = np.cos(angle)
    sinangle = np.sin(angle)
    xx[0] = sourcex - centerx
    yy[0] = sourcey - centery
    for i in range(number_apertures-1):
        if sens == 'clock':
            xx[i+1] = cosangle*xx[i] + sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] - sinangle*xx[i] 
        elif sens == 'counterclock':
            xx[i+1] = cosangle*xx[i] - sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] + sinangle*xx[i]

    xx += centerx
    yy += centery
    rad = aperture*fwhm
    if exclude_negative_lobes:
        xx = np.concatenate(([xx[0]], xx[2:-1]))
        yy = np.concatenate(([yy[0]], yy[2:-1]))

    sigmas = np.zeros_like(yy)
    for i in range(1,len(sigmas)):
        sigmas[i] = np.std(get_circle(array, rad, cy=yy[i], cx=xx[i], 
                                      mode="val"), ddof=1)

    sigmas = sigmas[1:] # do not consider companion itself
    sigma = np.mean(sigmas)

    if full_output:
        return sigmas, sigma
    else:
        return sigma


#! /usr/bin/env python

"""
Module with NMF algorithm in concentric annuli for ADI/RDI.
"""

__author__ = 'Valentin Christiaens'
__all__ = ['nmf_annular']

import numpy as np
from multiprocessing import cpu_count
from sklearn.decomposition import NMF
from ..preproc import cube_derotate, cube_collapse, check_pa_vector
from ..preproc.derotation import _find_indices_adi, _define_annuli
from ..var import get_annulus_segments, matrix_scaling
from ..conf import timing, time_ini
from ..conf.utils_conf import pool_map, iterable


def nmf_annular(cube, angle_list, cube_ref=None, radius_int=0, fwhm=4, asize=4, 
                n_segments=1, delta_rot=(0.1, 1), ncomp=1, init_svd='nndsvd', 
                nproc=1, min_frames_lib=2, max_frames_lib=200, scaling=None, 
                imlib='opencv', interpolation='lanczos4', collapse='median', 
                full_output=False, verbose=True, theta_init=0, weights=None, 
                cube_sig=None, handle_neg='mask', max_iter=1000, 
                random_state=None, **kwargs):
    """ Non Negative Matrix Factorization in concentric annuli, for ADI/RDI 
    sequences. Alternative to the annular ADI-PCA processing that does not rely 
    on SVD or ED for obtaining a low-rank approximation of the datacube. 
    This function embeds the scikit-learn NMF algorithm solved through either 
    the coordinate descent or the multiplicative update method.
    
    Parameters
    ----------
    cube : numpy ndarray, 3d
        Input cube.
    angle_list : numpy ndarray, 1d
        Corresponding parallactic angle for each frame.   
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the
        central circular region is discarded.
    fwhm : float, optional
        Size of the FHWM in pixels. Default is 4.
    asize : float, optional
        The size of the annuli, in pixels.
    n_segments : int or list of ints or 'auto', optional
        The number of segments for each annulus. When a single integer is given
        it is used for all annuli. When set to 'auto', the number of segments is
        automatically determined for every annulus, based on the annulus width.
    delta_rot : float or tuple of floats, optional
        Factor for adjusting the parallactic angle threshold, expressed in
        FWHM. Default is 1 (excludes 1 FHWM on each side of the considered
        frame). If a tuple of two floats is provided, they are used as the lower
        and upper intervals for the threshold (grows linearly as a function of
        the separation). !!! Important: this is used even if a reference cube 
        is provided for RDI. This is to allow ARDI (PCA library built from both
        science and reference cubes). If you want to do pure RDI, set delta_rot 
        to an arbitrarily high value such that the condition is never fulfilled
        for science frames to make it in the PCA library.
    ncomp : int, optional
        How many components are used as for low-rank approximation of the 
        datacube.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    max_iter : int optional
        The number of iterations for the coordinate descent solver.
    random_state : int or None, optional
        Controls the seed for the Pseudo Random Number generator.
    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.
    source_xy : tuple of int, optional
        For ADI-PCA, this triggers a frame rejection in the PCA library, with
        ``source_xy`` as the coordinates X,Y of the center of the annulus where
        the PA criterion is estimated. When ``ncomp`` is a tuple, a PCA grid is
        computed and the S/Ns (mean value in a 1xFWHM circular aperture) of the
        given (X,Y) coordinates are computed.
    delta_rot : int, optional
        Factor for tunning the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1xFHWM on each side of the considered frame).
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default value is 4. 
    init_svd: str, optional {'nnsvd','nnsvda','random'}
        Method used to initialize the iterative procedure to find H and W.
        'nndsvd': non-negative double SVD recommended for sparseness
        'nndsvda': NNDSVD where zeros are filled with the average of cube; 
        recommended when sparsity is not desired
        'random': random initial non-negative matrix
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    kwargs 
        Additional arguments for scikit-learn NMF algorithm. See:
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html             
             
    Returns
    -------
    If full_output is False the final frame is returned. If True the algorithm
    returns the reshaped NMF components, the reconstructed cube, the residuals,
    the residuals derotated and the final frame.     
    
    """
    if verbose:
        global start_time
        start_time = time_ini()
        
    array = cube
    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array')
    if array.shape[0] != angle_list.shape[0]:
        raise TypeError('Input vector or parallactic angles has wrong length')

    n, y, _ = array.shape

    angle_list = check_pa_vector(angle_list)
    n_annuli = int((y / 2 - radius_int) / asize)

    if isinstance(delta_rot, tuple):
        delta_rot = np.linspace(delta_rot[0], delta_rot[1], num=n_annuli)
    elif isinstance(delta_rot, (int, float)):
        delta_rot = [delta_rot] * n_annuli

    if isinstance(n_segments, int):
        n_segments = [n_segments for _ in range(n_annuli)]
    elif n_segments == 'auto':
        n_segments = list()
        n_segments.append(2)  # for first annulus
        n_segments.append(3)  # for second annulus
        ld = 2 * np.tan(360 / 4 / 2) * asize
        for i in range(2, n_annuli):  # rest of annuli
            radius = i * asize
            ang = np.rad2deg(2 * np.arctan(ld / (2 * radius)))
            n_segments.append(int(np.ceil(360 / ang)))

    if verbose:
        msg = 'N annuli = {}, FWHM = {:.3f}'
        print(msg.format(n_annuli, fwhm))
        print('PCA per annulus (or annular sectors):')

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = cpu_count() // 2

    # how to handle negative values
    if handle_neg=='null':
        array[np.where(array<0)]=0
    elif handle_neg=='subtr_min':
        array -= np.amin(array)
    elif not handle_neg=='mask':
        raise ValueError("Mode to handle neg. pixels not recognized")

    # The annuli are built, and the corresponding PA thresholds for frame
    # rejection are calculated (at the center of the annulus)
    cube_out = np.zeros_like(array)
    cube_recon = np.zeros_like(array)
    H_comps = np.zeros([ncomp,array.shape[1],array.shape[2]])
    if cube_ref is None:
        strict = False
    else:
        strict = True
    for ann in range(n_annuli):
        if isinstance(ncomp, tuple) or isinstance(ncomp, np.ndarray):
            if len(ncomp) == n_annuli:
                ncompann = ncomp[ann]
            else:
                raise TypeError('If `ncomp` is a tuple, it must match the '
                                'number of annuli')
        else:
            ncompann = ncomp

        n_segments_ann = n_segments[ann]
        res_ann_par = _define_annuli(angle_list, ann, n_annuli, fwhm,
                                     radius_int, asize, delta_rot[ann],
                                     n_segments_ann, verbose, strict)
        pa_thr, inner_radius, ann_center = res_ann_par
        indices = get_annulus_segments(array[0], inner_radius, asize,
                                       n_segments_ann, theta_init)
        # Library matrix is created for each segment and scaled if needed
        for j in range(n_segments_ann):
            yy = indices[j][0]
            xx = indices[j][1]
            if handle_neg=='mask':
                npts = range(len(yy))
                if cube_sig is not None:
                    yp = [yy[i] for i in npts if np.amin(array[:, yy[i], xx[i]]-np.abs(cube_sig[:, yy[i], xx[i]]))>0]
                    xp = [xx[i] for i in npts if np.amin(array[:, yy[i], xx[i]]-np.abs(cube_sig[:, yy[i], xx[i]]))>0]
                else:
                    yp = [yy[i] for i in npts if np.amin(array[:, yy[i], xx[i]])>0]
                    xp = [xx[i] for i in npts if np.amin(array[:, yy[i], xx[i]])>0]
                yy = tuple(yp)
                xx = tuple(xp)
            matrix_segm = array[:, yy, xx]  # shape [nframes x npx_segment]
            matrix_segm = matrix_scaling(matrix_segm, scaling)
            if cube_ref is not None:
                matrix_segm_ref = cube_ref[:, yy, xx]
                matrix_segm_ref = matrix_scaling(matrix_segm_ref, scaling)
            else:
                matrix_segm_ref = None
            if cube_sig is not None:
                matrix_sig_segm = cube_sig[:, yy, xx]
            else:
                matrix_sig_segm = None

            res = pool_map(nproc, do_nmf_patch, matrix_segm, iterable(range(n)),
                           angle_list, fwhm, pa_thr, ann_center, ncompann, 
                           max_iter, random_state, init_svd, min_frames_lib, 
                           max_frames_lib, matrix_segm_ref, matrix_sig_segm,
                           handle_neg)

            res = np.array(res, dtype=object)
            residuals = np.array(res[:, 0])
            # ncomps = res[:, 1]
            # nfrslib = res[:, 2]
            recon = np.array(res[:, 1])
            H = np.array(res[:, 2])
            for fr in range(n):
                cube_out[fr][yy, xx] = residuals[fr]
                cube_recon[fr][yy, xx] = recon[fr]
            for pp in range(ncomp):
                H_comps[pp][yy, xx] = H[0][pp] # just save H inferred for fr=0

        if verbose == 2:
            timing(start_time)

    # Cube is derotated according to the parallactic angle and collapsed
    cube_der = cube_derotate(cube_out, angle_list, imlib=imlib,
                             interpolation=interpolation)
    frame = cube_collapse(cube_der, mode=collapse, w=weights)
    if verbose:
        print('Done derotating and combining.')
        timing(start_time)
    if full_output:
        return cube_out, cube_der, cube_recon, H_comps, frame
    else:
        return frame

    
def do_nmf_patch(matrix, frame, angle_list, fwhm, pa_threshold, ann_center,
                 ncomp, max_iter, random_state, init_svd, min_frames_lib, 
                 max_frames_lib, matrix_ref, matrix_sig_segm, handle_neg,
                 **kwargs):
    """ Solves the NMF for each frame patch (small matrix). For each frame we
    find the frames to be rejected depending on the amount of rotation. The
    library is also truncated on the other end (frames too far or which have
    rotated more) which are more decorrelated to keep the computational cost
    lower. This truncation is done on the annuli after 10*FWHM and the goal is
    to keep min(num_frames/2, 200) in the library.
    """

    # Note: blocks below allow the possibility of ARDI
    if pa_threshold != 0:
        indices_left = _find_indices_adi(angle_list, frame, pa_threshold, 
                                         truncate=True, 
                                         max_frames=max_frames_lib)
        msg = 'Too few frames left in the PCA library. '
        msg += 'Accepted indices length ({:.0f}) less than {:.0f}. '
        msg += 'Try decreasing either delta_rot or min_frames_lib.'
        try:
            if matrix_sig_segm is not None:
                data_ref = matrix[indices_left]-matrix_sig_segm[indices_left]
            else:
                data_ref = matrix[indices_left]
        except IndexError:
            if matrix_ref is None:
                raise RuntimeError(msg.format(0, min_frames_lib))
            data_ref = None

        if data_ref.shape[0] < min_frames_lib and matrix_ref is None:
            raise RuntimeError(msg.format(data_ref.shape[0], min_frames_lib))
    elif pa_threshold == 0:
        if matrix_sig_segm is not None:
            data_ref = matrix-matrix_sig_segm
        else:
            data_ref = matrix
    if matrix_ref is not None:
        if data_ref is not None:
            data_ref = np.vstack((matrix_ref, data_ref))
        else:
            data_ref = matrix_ref

    # to avoid bug, just consider positive values
    if np.median(data_ref)<0:
        raise ValueError("Mostly negative values in the cube")
    else:
        # how to handle negative values
        if handle_neg=='null':
            data_ref[np.where(data_ref<0)]=0
        elif handle_neg=='subtr_min':
            data_ref -= np.amin(data_ref)
        else: #'mask'
            zp = np.nonzero(np.amin(data_ref,axis=0)>0)
        
    solver = 'mu'
    # if init_svd == 'nndsvda':
    #     solver = 'mu'
    # else:
    #     solver = 'cd'
    mod = NMF(n_components=ncomp, alpha=0, solver=solver, init=init_svd, 
              max_iter=max_iter, random_state=random_state, **kwargs)  

    curr_frame = matrix[frame]  # current frame
    if matrix_sig_segm is not None:
        curr_frame_emp = matrix[frame]-matrix_sig_segm[frame]
    else:
        curr_frame_emp = curr_frame.copy()
    # how to handle negative values
    if handle_neg=='null':
        curr_frame_emp[np.where(curr_frame_emp<0)]=0
    elif handle_neg=='subtr_min':
        curr_frame_emp -= np.amin(curr_frame_emp)
    else: #'mask'
        zzp = np.nonzero(curr_frame_emp>0)
        pos_p = np.intersect1d(zp[0],zzp[0])
        curr_frame_emp = curr_frame_emp[pos_p]    
        data_ref = data_ref[:,pos_p]
        
    H = mod.fit(data_ref).components_
    W = mod.transform(curr_frame_emp[np.newaxis,...])
    reconstructed = np.dot(W, H)
    # if masked neg values, reshape
    if handle_neg == 'mask': #'mask'
        recon = np.zeros(matrix.shape[1])
        recon[pos_p] = reconstructed
        reconstructed = recon.copy()
        H_tmp = np.zeros([ncomp,matrix.shape[1]])
        for pp in range(ncomp):
            H_tmp[pp,pos_p] = H[pp]
        H = H_tmp.copy()   
    residuals = curr_frame - reconstructed
    return residuals, reconstructed, H
    #return residuals, V.shape[0], data_ref.shape[0]    


#! /usr/bin/env python

"""
Module for image statistics.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_histo_stats',
           'frame_average_radprofile']

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from ..var import frame_center
from ..conf.utils_conf import check_array, vip_figsize


def frame_average_radprofile(frame, sep=1, init_rad=None, plot=True):
    """ Calculates the average radial profile of an image.

    Parameters
    ----------
    frame : numpy ndarray
        Input image or 2d array.
    sep : int, optional
        The average radial profile is recorded every ``sep`` pixels.
    plot : bool, optional
        If True the profile is plotted.

    Returns
    -------
    df : dataframe
        Pandas dataframe with the radial profile and distances.

    Notes
    -----
    https://stackoverflow.com/questions/21242011/most-efficient-way-to-calculate-radial-profile
    https://stackoverflow.com/questions/48842320/what-is-the-best-way-to-calculate-radial-average-of-the-image-with-python
    https://github.com/keflavich/image_tools/blob/master/image_tools/radialprofile.py

    """
    check_array(frame, dim=2)
    cy, cx = frame_center(frame)

    if init_rad is None:
        init_rad = 1
    x, y = np.indices((frame.shape))
    r = np.sqrt((x - cx) ** 2 + (y - cy) ** 2)
    r = r.astype(int)
    tbin = np.bincount(r.ravel(), frame.ravel())
    nr = np.bincount(r.ravel())
    radprofile = tbin / nr

    radists = np.arange(init_rad + 1, int(cy), sep) - 1
    radprofile_radists = radprofile[radists]
    nr_radists = nr[radists]
    df = pd.DataFrame({'rad': radists, 'radprof': radprofile_radists,
                       'npx': nr_radists})

    if plot:
        plt.figure(figsize=vip_figsize)
        plt.plot(radists, radprofile_radists, '.-', alpha=0.6)
        plt.grid(which='both', alpha=0.4)
        plt.xlabel('Pixels')
        plt.ylabel('Counts')
        plt.minorticks_on()
        plt.xlim(0)

    return df


def frame_histo_stats(image_array, plot=True):
    """Plots a frame with a colorbar, its histogram and some statistics: mean,
    median, maximum, minimum and standard deviation values.  
    
    Parameters
    ----------
    image_array : numpy ndarray
        The input frame.  
    plot : bool, optional
        If True plots the frame and the histogram with the values.
        
    Return
    ------
    mean : float
        Mean value of array.
    median : float
        Median value of array.
    std : float
        Standard deviation of array.
    maxim : int or float
        Maximum value.
    minim : int or float
        Minimum value.
        
    """
    vector = image_array.flatten()
    mean = vector.mean()
    median = np.median(vector)
    maxim = vector.max()
    minim = vector.min()
    std = vector.std()
   
    if plot:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 4))
        ax0, ax1 = axes.flat
        bins = int(np.sqrt(vector.shape[0]))
        txt = 'Mean = {:.3f}\n'.format(mean) + \
              'Median = {:.3f}\n'.format(median) +\
              'Stddev = {:.3f}\n'.format(std) +\
              'Max = {:.3f}\n'.format(maxim) +\
              'Min = {:.3f}\n\n'.format(minim)

        ax0.imshow(image_array, interpolation="nearest", origin ="lower",
                   cmap='viridis')
        ax0.set_title('Frame')
        ax0.grid('off')

        ax1.hist(vector, bins=bins, label=txt, alpha=0.5, histtype='stepfilled')
        ax1.set_yscale('log')
        ax1.set_title('Histogram')
        ax1.text(0.98, 0.98, txt, transform=ax1.transAxes, fontsize=12,
                 verticalalignment='top', horizontalalignment='right')
        plt.show()
        
    return mean, median, std, maxim, minim




#! /usr/bin/env python

"""
Module with fake disk injection functions.
"""

__author__ = 'Julien Milli @ ESO, Valentin Christiaens @ ULg/UChile'
__all__ = ['cube_inject_fakedisk',
           'cube_inject_trace']

import numpy as np
from scipy import signal
from ..preproc import cube_derotate, frame_shift
from ..var import frame_center


def cube_inject_fakedisk(fakedisk, angle_list, psf=None, imlib='opencv',
                         interpolation='lanczos4', cxy=None, nproc=1,
                         border_mode='constant'):
    """
    Rotates an ADI cube to a common north given a vector with the corresponding
    parallactic angles for each frame of the sequence. By default bicubic
    interpolation is used (opencv).

    Parameters
    ----------
    fakedisk : numpy ndarray
        Input image of a fake disc
    angle_list : list
        Vector containing the parallactic angles.
    psf : (optional) the PSF to convolve the disk image with. It can be a
        small numpy.ndarray (we advise to use odd sizes to make sure the center
        s not shifted through the convolution). It forces normalization of the
        PSF to preserve the flux. It can also be a float representing
        the FWHM of the gaussian to be used for convolution.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    cxy : tuple of int, optional
        Coordinates X,Y  of the point with respect to which the rotation will be
        performed. By default the rotation is done with respect to the center
        of the frames, as it is returned by the function
        vip_hci.var.frame_center.
    nproc : int, optional
        Whether to rotate the frames in the sequence in a multi-processing
        fashion. Only useful if the cube is significantly large (frame size and
        number of frames).
    border_mode : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.

    Returns
    -------
    fakedisk_cube : numpy ndarray
        Resulting cube with the fake disc inserted at the correct angles and
        convolved with the psf if a psf was provided.

    Notes
    -----
    .. code-block:: python

        import numpy as np

        fakedisk = np.zeros((200,200))
        fakedisk[:,99:101] = 1
        angle_list = np.arange(10)
        c = create_fakedisk_cube(fakedisk, angle_list, psf=None, imlib='opencv',
                                 interpolation='lanczos4',cxy=None, nproc=1,
                                 border_mode='constant')

    """
    if not fakedisk.ndim == 2:
        raise TypeError('Fakedisk is not a frame or a 2d array.')
    if not angle_list.ndim == 1:
        raise TypeError('Input parallactic angle is not a 1d array')
    nframes = len(angle_list)
    ny, nx = fakedisk.shape
    fakedisk_cube = np.repeat(fakedisk[np.newaxis, :, :], nframes, axis=0)
    fakedisk_cube = cube_derotate(fakedisk_cube, angle_list, imlib=imlib,
                                  interpolation=interpolation, cxy=cxy,
                                  nproc=nproc, border_mode=border_mode)

    if psf is not None:
        if isinstance(psf, np.ndarray):
            if psf.ndim != 2:
                raise TypeError('Input PSF is not a frame or 2d array.')
            if np.abs(np.sum(psf)-1) > 1e-4:
                print('Warning the PSF is not normalized to a total of 1. '
                      'Normalization was forced.')
                psf = psf/np.sum(psf)
        elif isinstance(psf, (int, float)):
            # assumes psf is equal to the FWHM of the PSF. We create a synthetic
            # PSF in that case
            # with a size of 2 times the FWHM.
            psf_size = 2*int(np.round(psf))+1  # to make sure this is odd.
            xarrray, yarray = np.meshgrid(np.arange(-(psf_size//2),
                                                    psf_size//2+1),
                                          np.arange(-(psf_size//2),
                                                    psf_size//2+1))
            d = np.sqrt(xarrray**2+yarray**2)
            sigma = psf/(2*np.sqrt(2*np.log(2)))
            psf = np.exp(-(d**2 / (2.0*sigma**2)))
            psf = psf/np.sum(psf)
        else:
            raise TypeError('The type of the psf is unknown. '
                            'create_fakedisk_cube accepts ndarray, int or '
                            'float.')
        for i in range(nframes):
            fakedisk_cube[i, :, :] = signal.convolve2d(fakedisk_cube[i, :, :],
                                                       psf, mode='same')
    return fakedisk_cube


def cube_inject_trace(array, psf_template, angle_list, flevel, rad_dists, theta, 
                      plsc=0.01225, n_branches=1, imlib='opencv', verbose=True):
    """ Injects fake companions along a trace, such as a spiral. The trace is 
    provided by 2 arrays corresponding to the polar coordinates where the 
    companions will be located in the final derotated frame.
    Note: for a continuous-looking trace, and for an easier scaling using 
    parameter 'flevel', it is recommended to separate the points of the trace
    by a distance of FWHM/2.
    
    Parameters
    ----------
    array : numpy ndarray
        Input 3D cube in which the extended feature is injected.
    psf_template : numpy ndarray 
        2d array with the normalized psf template. It should have an odd shape.
        It is recommended to run the function psf_norm to get a proper PSF
        template.
    flevel : float
        Flux at which the fake companions are injected into the cube along the 
        trace.
    rad_dists : list or array 1d
        Vector of radial distances where the trace is to be injected.
    theta : list or array 1d
        Vector of angles (deg) where the trace is to be injected (trigonometric
        angles, NOT PA East from North).
    plsc : float, opt
        Value of the plate scale in arcsec/pixel (optional, will only be used 
        if verbose is True).
    n_branches : int, optional
        Number of azimutal branches on which the trace is injected.
    imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
        Library or method used for image operations (shifts). Opencv is the
        default for being the fastest.
    verbose : {True, False}, bool optional
        If True prints out additional information. 
    
    Returns
    -------
    array_out : numpy ndarray
        Output array with the injected fake companions.
        
    """
    if not array.ndim==3: 
        raise TypeError('Array is not a cube or 3d array')
    
    ceny, cenx = frame_center(array[0])
    ceny = int(ceny)
    cenx = int(cenx)
    rad_dists = np.array(rad_dists)
    if not rad_dists[-1]<array[0].shape[0]/2.:
        msg = 'rad_dists last location is at the border (or outside) of the field'
        raise ValueError(msg)
    
    size_fc = psf_template.shape[0]
    nframes, ny, nx = array.shape
    n_fc_rad = rad_dists.shape[0]

    w = int(np.floor(size_fc/2.))

    array_out = np.zeros_like(array)
    for fr in range(nframes):
        tmp = np.zeros_like(array[0])
        for branch in range(n_branches):
            for i in range(n_fc_rad):
                fc_fr = np.zeros_like(array[0], dtype=np.float64)
                ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta[i])
                rad = rad_dists[i]                                         
                y = rad * np.sin(ang - np.deg2rad(angle_list[fr]))
                x = rad * np.cos(ang - np.deg2rad(angle_list[fr]))
                # deal with exceptions
                y0_fr = max(0,ceny+(int(y)-w))
                y0_psf = max(0,-(ceny+int(y)-w))
                x0_fr = max(0,cenx+(int(x)-w))
                x0_psf = max(0,-(cenx+int(x)-w))
                yn_fr = min(ny,ceny+(int(y)+w+1))     
                yn_psf = min(size_fc,size_fc-(ceny+(int(y)+w+1)-ny))        
                xn_fr = min(nx,cenx+(int(x)+w+1))     
                xn_psf = min(size_fc,size_fc-(cenx+(int(x)+w+1)-nx))     
                if x > 0:
                    mod_x = x%1.
                else:
                    mod_x = (x%1.)-1
                if y > 0:
                    mod_y = y%1.
                else:
                    mod_y = (y%1.)-1
                try:
                    psf_tmp = flevel*psf_template[y0_psf:yn_psf,x0_psf:xn_psf]
                    fc_fr[y0_fr:yn_fr, x0_fr:xn_fr] = frame_shift(psf_tmp, 
                                                                  mod_y, mod_x, 
                                                                  imlib=imlib)
                except:
                    raise TypeError('Problem with the coordinates of the trace')
                tmp += fc_fr
        array_out[fr] = array[fr] + tmp
    
    if verbose:
        for branch in range(n_branches):
            print('Branch '+str(branch+1)+':')
            for i in range(n_fc_rad):
                ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta[i])
                posy = rad_dists[i] * np.sin(ang) + ceny
                posx = rad_dists[i] * np.cos(ang) + cenx
                rad_arcs = rad_dists[i]*plsc
                msg ='\t(X,Y)=({:.2f}, {:.2f}) at {:.2f} arcsec ({:.2f} pxs)'
                print(msg.format(posx, posy, rad_arcs, rad_dists[i]))
        
    return array_out


#! /usr/bin/env python

"""
Full-frame PCA algorithm for ADI, ADI+RDI and ADI+mSDI (IFS data) cubes.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['pca']

import numpy as np
from multiprocessing import cpu_count
from .svd import svd_wrapper, SVDecomposer
from .utils_pca import pca_incremental, pca_grid
from ..preproc.derotation import _find_indices_adi, _compute_pa_thresh
from ..preproc import cube_rescaling_wavelengths as scwave
from ..preproc import (cube_derotate, cube_collapse, check_pa_vector,
                       check_scal_vector, cube_crop_frames, 
                       cube_subtract_sky_pca)
from ..conf import (timing, time_ini, check_enough_memory, Progressbar,
                    check_array)
from ..conf.utils_conf import pool_map, iterable
from ..var import (frame_center, dist, prepare_matrix, reshape_matrix,
                   cube_filter_lowpass, mask_circle)
from ..stats import descriptive_stats


def pca(cube, angle_list, cube_ref=None, scale_list=None, ncomp=1,
        svd_mode='lapack', scaling=None, mask_center_px=None, source_xy=None,
        delta_rot=1, fwhm=4, adimsdi='single', crop_ifs=True, imlib='opencv',
        imlib2='opencv', interpolation='lanczos4', collapse='median', 
        ifs_collapse_range='all', mask_rdi=None, check_memory=True, batch=None, 
        nproc=1, full_output=False, verbose=True, weights=None, conv=False,
        cube_sig=None):
    """ Algorithm where the reference PSF and the quasi-static speckle pattern
    are modeled using Principal Component Analysis. Depending on the input
    parameters this PCA function can work in ADI, RDI or SDI (IFS data) mode.

    ADI: the target ``cube`` itself is used to learn the PCs and to obtain a
    low-rank approximation model PSF (star + speckles). Both `cube_ref`` and
    ``scale_list`` must be None. The full-frame ADI-PCA implementation is based
    on Soummer et al. 2012 (http://arxiv.org/abs/1207.4197) and Amara & Quanz
    2012 (http://arxiv.org/abs/1207.6637). If ``batch`` is provided then the
    cube if processed with incremental PCA as described in Gomez Gonzalez et al.
    2017 (https://arxiv.org/abs/1705.06184).

    ADI + RDI: if a reference cube is provided (``cube_ref``), its PCs are used
    to reconstruct the target frames to obtain the model PSF (star + speckles).

    ADI + SDI (IFS data): if a scaling vector is provided (``scale_list``) and
    the cube is a 4d array [# channels, # adi-frames, Y, X], its assumed it
    contains several multi-spectral ADI frames. A single or two stages PCA can
    be performed, depending on ``adimsdi``.

    Parameters
    ----------
    cube : str or numpy ndarray, 3d or 4d
        Input cube (ADI or ADI+mSDI). If a string is given, it must correspond
        to the path to the fits file to be opened in memmap mode (for PCA
        incremental of ADI 3d cubes).
    angle_list : numpy ndarray, 1d
        Corresponding parallactic angle for each frame.
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    scale_list : numpy ndarray, 1d, optional
        Scaling factors in case of IFS data (ADI+mSDI cube). Usually, the
        scaling factors are the central channel wavelength divided by the
        shortest wavelength in the cube (more thorough approaches can be used
        to get the scaling factors). This scaling factors are used to re-scale
        the spectral channels and align the speckles.
    ncomp : int, float or tuple of int/None, optional
        How many PCs are used as a lower-dimensional subspace to project the
        target frames.

        * ADI (``cube`` is a 3d array): if an int is provided, ``ncomp`` is the
          number of PCs extracted from ``cube`` itself. If ``ncomp`` is a float
          in the interval (0, 1] then it corresponds to the desired cumulative
          explained variance ratio (the corresponding number of components is
          estimated). If ``ncomp`` is a tuple of two integers, then it
          corresponds to an interval of PCs in which final residual frames are
          computed (optionally, if a tuple of 3 integers is passed, the third
          value is the step). When``source_xy`` is not None, then the S/Ns
          (mean value in a 1xFWHM circular aperture) of the given
          (X,Y) coordinates are computed.

        * ADI+RDI (``cube`` and ``cube_ref`` are 3d arrays): ``ncomp`` is the
          number of PCs obtained from ``cube_ref``. If ``ncomp`` is a tuple,
          then it corresponds to an interval of PCs (obtained from ``cube_ref``)
          in which final residual frames are computed. If ``source_xy`` is not
          None, then the S/Ns (mean value in a 1xFWHM circular aperture) of the
          given (X,Y) coordinates are computed.

        * ADI+mSDI (``cube`` is a 4d array and ``adimsdi="single"``): ``ncomp``
          is the number of PCs obtained from the whole set of frames
          (n_channels * n_adiframes). If ``ncomp`` is a float in the interval
          (0, 1] then it corresponds to the desired CEVR, and the corresponding
          number of components will be estimated. If ``ncomp`` is a tuple, then
          it corresponds to an interval of PCs in which final residual frames
          are computed. If ``source_xy`` is not None, then the S/Ns (mean value
          in a 1xFWHM circular aperture) of the given (X,Y) coordinates are
          computed.

        * ADI+mSDI  (``cube`` is a 4d array and ``adimsdi="double"``): ``ncomp``
          must be a tuple, where the first value is the number of PCs obtained
          from each multi-spectral frame (if None then this stage will be
          skipped and the spectral channels will be combined without
          subtraction); the second value sets the number of PCs used in the
          second PCA stage, ADI-like using the residuals of the first stage (if
          None then the second PCA stage is skipped and the residuals are
          de-rotated and combined).

    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy',
        'randcupy', 'pytorch', 'eigenpytorch', 'randpytorch'}, str optional
        Switch for the SVD method/library to be used.

        ``lapack``: uses the LAPACK linear algebra library through Numpy
        and it is the most conventional way of computing the SVD
        (deterministic result computed on CPU).

        ``arpack``: uses the ARPACK Fortran libraries accessible through
        Scipy (computation on CPU).

        ``eigen``: computes the singular vectors through the
        eigendecomposition of the covariance M.M' (computation on CPU).

        ``randsvd``: uses the randomized_svd algorithm implemented in
        Sklearn (computation on CPU).

        ``cupy``: uses the Cupy library for GPU computation of the SVD as in
        the LAPACK version. `

        `eigencupy``: offers the same method as with the ``eigen`` option
        but on GPU (through Cupy).

        ``randcupy``: is an adaptation of the randomized_svd algorithm,
        where all the computations are done on a GPU (through Cupy). `

        `pytorch``: uses the Pytorch library for GPU computation of the SVD.

        ``eigenpytorch``: offers the same method as with the ``eigen``
        option but on GPU (through Pytorch).

        ``randpytorch``: is an adaptation of the randomized_svd algorithm,
        where all the linear algebra computations are done on a GPU
        (through Pytorch).

    scaling : {None, "temp-mean", spat-mean", "temp-standard",
        "spat-standard"}, None or str optional
        Pixel-wise scaling mode using ``sklearn.preprocessing.scale``
        function. If set to None, the input matrix is left untouched. Otherwise:

        ``temp-mean``: temporal px-wise mean is subtracted.

        ``spat-mean``: spatial mean is subtracted.

        ``temp-standard``: temporal mean centering plus scaling pixel values
        to unit variance. HIGHLY RECOMMENDED FOR ASDI AND RDI CASES!

        ``spat-standard``: spatial mean centering plus scaling pixel values
        to unit variance.

    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.
    source_xy : tuple of int, optional
        For ADI-PCA, this triggers a frame rejection in the PCA library, with
        ``source_xy`` as the coordinates X,Y of the center of the annulus where
        the PA criterion is estimated. When ``ncomp`` is a tuple, a PCA grid is
        computed and the S/Ns (mean value in a 1xFWHM circular aperture) of the
        given (X,Y) coordinates are computed.
    delta_rot : int, optional
        Factor for tunning the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1xFHWM on each side of the considered frame).
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default value is 4. 
    adimsdi : {'single', 'double'}, str optional
        Changes the way the 4d cubes (ADI+mSDI) are processed. Basically it
        determines whether a single or double pass PCA is going to be computed.

        ``single``: the multi-spectral frames are rescaled wrt the largest
        wavelength to align the speckles and all the frames (n_channels *
        n_adiframes) are processed with a single PCA low-rank approximation.

        ``double``: a first stage is run on the rescaled spectral frames, and a
        second PCA frame is run on the residuals in an ADI fashion.

    crop_ifs: bool, optional
        [adimsdi='single'] If True cube is cropped at the moment of frame
        rescaling in wavelength. This is recommended for large FOVs such as the
        one of SPHERE, but can remove significant amount of information close to
        the edge of small FOVs (e.g. SINFONI).
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    imlib2 : str, optional
        See the documentation of the ``vip_hci.preproc.cube_rescaling_wavelengths`` 
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    ifs_collapse_range: str 'all' or tuple of 2 int
        If a tuple, it should contain the first and last channels where the mSDI 
        residual channels will be collapsed (by default collapses all channels).
    mask_rdi: 2d numpy array, opt
        If provided, this binary mask will be used either in RDI mode or in 
        ADI+mSDI (2 steps) mode. The projection coefficients for the principal 
        components will be found considering the area covered by the mask 
        (useful to avoid self-subtraction in presence of bright disc signal)
    check_memory : bool, optional
        If True, it checks that the input cube is smaller than the available
        system memory.
    batch : None, int or float, optional
        When it is not None, it triggers the incremental PCA (for ADI and
        ADI+mSDI cubes). If an int is given, it corresponds to the number of
        frames in each sequential mini-batch. If a float (0, 1] is given, it
        corresponds to the size of the batch is computed wrt the available
        memory in the system.
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of
        processes will be set to (cpu_count()/2). Defaults to ``nproc=1``.
    full_output: bool, optional
        Whether to return the final median combined image only or with other
        intermediate arrays.
    verbose : bool, optional
        If True prints intermediate info and timing.
    weights: 1d numpy array or list, optional
        Weights to be applied for a weighted mean. Need to be provided if 
        collapse mode is 'wmean'.
    cube_sig: numpy ndarray, opt
        Cube with estimate of significant authentic signals. If provided, this
        will subtracted before projecting cube onto reference cube.
        
    Returns
    -------
    frame : numpy ndarray
        2D array, median combination of the de-rotated/re-scaled residuals cube.
        Always returned.
    pcs : numpy ndarray
        [full_output=True, source_xy=None] Principal components. Valid for
        ADI cubes (3D). This is also returned when ``batch`` is not None
        (incremental PCA).
    recon_cube, recon : numpy ndarray
        [full_output=True] Reconstructed cube. Valid for ADI cubes (3D).
    residuals_cube : numpy ndarray
        [full_output=True] Residuals cube. Valid for ADI cubes (3D).
    residuals_cube_ : numpy ndarray
        [full_output=True] Derotated residuals cube. Valid for ADI cubes (3D).
    residuals_cube_channels : numpy ndarray
        [full_output=True, adimsdi='double'] Residuals for each multi-spectral
        cube. Valid for ADI+mSDI (4D) cubes.
    residuals_cube_channels_ : numpy ndarray
        [full_output=True, adimsdi='double'] Derotated final residuals. Valid
        for ADI+mSDI (4D) cubes.
    cube_allfr_residuals : numpy ndarray
        [full_output=True, adimsdi='single']  Residuals cube (of the big cube
        with channels and time processed together). Valid for ADI+mSDI (4D)
        cubes.
    cube_adi_residuals : numpy ndarray
        [full_output=True, adimsdi='single'] Residuals cube (of the big cube
        with channels and time processed together) after de-scaling the wls.
        Valid for ADI+mSDI (4D).
    medians : numpy ndarray
        [full_output=True, source_xy=None] This is also returned when ``batch``
        is not None (incremental PCA).
    final_residuals_cube : numpy ndarray
        [ncomp is tuple] The residual final PCA frames for a grid a PCs.


    """
    start_time = time_ini(verbose)

    if batch is None:
        check_array(cube, (3, 4), msg='cube')
    else:
        if not isinstance(cube, (str, np.ndarray)):
            raise TypeError('`cube` must be a numpy (3d or 4d) array or a str '
                            'with the full path on disk')

    # checking memory (if in-memory numpy array is provided)
    if not isinstance(cube, str):
        input_bytes = cube_ref.nbytes if cube_ref is not None else cube.nbytes
        mem_msg = 'Set check_memory=False to override this memory check or ' \
                  'set `batch` to run incremental PCA (valid for ADI or ' \
                  'ADI+mSDI single-pass)'
        check_enough_memory(input_bytes, 1.0, raise_error=check_memory,
                            error_msg=mem_msg, verbose=verbose)

    if nproc is None:
        nproc = cpu_count() // 2        # Hyper-threading doubles the # of cores

    # ADI + mSDI. Shape of cube: (n_channels, n_adi_frames, y, x)
    if scale_list is not None: #isinstance(cube, np.ndarray) and cube.ndim == 4:
        if adimsdi == 'double':
            res_pca = _adimsdi_doublepca(cube, angle_list, scale_list, ncomp,
                                         scaling, mask_center_px, svd_mode,
                                         imlib, imlib2, interpolation, collapse, 
                                         ifs_collapse_range, verbose, start_time, 
                                         nproc, weights, fwhm, conv, mask_rdi,
                                         cube_sig)
            residuals_cube_channels, residuals_cube_channels_, frame = res_pca
        elif adimsdi == 'single':
            res_pca = _adimsdi_singlepca(cube, angle_list, scale_list, ncomp,
                                         fwhm, source_xy, scaling,
                                         mask_center_px, svd_mode, imlib,
                                         imlib2, interpolation, collapse, 
                                         ifs_collapse_range, verbose, start_time, 
                                         crop_ifs, batch, full_output=True,
                                         weights=weights)
            if isinstance(ncomp, (int, float)):
                cube_allfr_residuals, cube_adi_residuals, frame = res_pca
            elif isinstance(ncomp, tuple):
                if source_xy is None:
                    final_residuals_cube, pclist = res_pca
                else:
                    final_residuals_cube, frame, table, _ = res_pca
        else:
            raise ValueError('`adimsdi` mode not recognized')

    # ADI + RDI
    elif cube_ref is not None:
        res_pca = _adi_rdi_pca(cube, cube_ref, angle_list, ncomp, scaling,
                               mask_center_px, svd_mode, imlib, interpolation,
                               collapse, verbose, start_time, weights, mask_rdi,
                               cube_sig)
        pcs, recon, residuals_cube, residuals_cube_, frame = res_pca

    # ADI. Shape of cube: (n_adi_frames, y, x)
    elif cube_ref is None:
        res_pca = _adi_pca(cube, angle_list, ncomp, batch, source_xy, delta_rot,
                           fwhm, scaling, mask_center_px, svd_mode, imlib,
                           interpolation, collapse, verbose, start_time, True,
                           weights, cube_sig)

        if batch is None:
            if source_xy is not None:
                # PCA grid, computing S/Ns
                if isinstance(ncomp, tuple):
                    if full_output:
                        final_residuals_cube, frame, table, _ = res_pca
                    else:
                        # returning only the optimal residual
                        final_residuals_cube = res_pca[1]
                # full-frame PCA with rotation threshold
                else:
                    recon_cube, residuals_cube, residuals_cube_, frame = res_pca
            else:
                # PCA grid
                if isinstance(ncomp, tuple):
                    final_residuals_cube, pclist = res_pca
                # full-frame standard PCA
                else:
                    pcs, recon, residuals_cube, residuals_cube_, frame = res_pca
        # full-frame incremental PCA
        else:
            frame, _, pcs, medians = res_pca

    else:
        raise RuntimeError('Only ADI, ADI+RDI and ADI+mSDI observing techniques'
                           ' are supported')

    # --------------------------------------------------------------------------
    # Returns for each case (ADI, ADI+RDI and ADI+mSDI) and combination of
    # parameters: full_output, source_xy, batch, ncomp
    # --------------------------------------------------------------------------
    if isinstance(cube, np.ndarray) and cube.ndim == 4:
        # ADI+mSDI double-pass PCA
        if adimsdi == 'double':
            if full_output:
                return frame, residuals_cube_channels, residuals_cube_channels_
            else:
                return frame

        elif adimsdi == 'single':
            # ADI+mSDI single-pass PCA
            if isinstance(ncomp, (float, int)):
                if full_output:
                    return frame, cube_allfr_residuals, cube_adi_residuals
                else:
                    return frame
            # ADI+mSDI single-pass PCA grid
            elif isinstance(ncomp, tuple):
                if source_xy is None and full_output:
                    return final_residuals_cube, pclist
                elif source_xy is None and not full_output:
                    return final_residuals_cube
                elif source_xy is not None and full_output:
                    return final_residuals_cube, frame, table
                elif source_xy is not None and not full_output:
                    return frame

    # ADI and ADI+RDI
    else:
        if (cube_ref is not None or source_xy is None) and full_output:
            # incremental PCA
            if batch is not None:
                return frame, pcs, medians
            else:
                # PCA grid
                if isinstance(ncomp, tuple):
                    return final_residuals_cube, pclist
                # full-frame standard PCA or ADI+RDI
                else:
                    return frame, pcs, recon, residuals_cube, residuals_cube_
        elif source_xy is not None and full_output:
            # PCA grid, computing S/Ns
            if isinstance(ncomp, tuple):
                return final_residuals_cube, frame, table
            # full-frame PCA with rotation threshold
            else:
                return frame, recon_cube, residuals_cube, residuals_cube_
        elif not full_output:
            # PCA grid
            if isinstance(ncomp, tuple):
                return final_residuals_cube
            # full-frame standard PCA or ADI+RDI
            else:
                return frame
        

def _adi_pca(cube, angle_list, ncomp, batch, source_xy, delta_rot, fwhm,
             scaling, mask_center_px, svd_mode, imlib, interpolation, collapse,
             verbose, start_time, full_output, weights=None, cube_sig=None):
    """ Handles the ADI PCA post-processing.
    """
    # Full/Single ADI processing, incremental PCA
    if batch is not None:
        result = pca_incremental(cube, angle_list, batch=batch, ncomp=ncomp,
                                 imlib=imlib, interpolation=interpolation,
                                 collapse=collapse, verbose=verbose,
                                 full_output=full_output, start_time=start_time,
                                 weights=weights)
        return result

    else:
        # Full/Single ADI processing
        n, y, x = cube.shape

        angle_list = check_pa_vector(angle_list)
        if not n == angle_list.shape[0]:
            raise ValueError("`angle_list` vector has wrong length. It must "
                             "equal the number of frames in the cube")

        if not isinstance(ncomp, (int, float, tuple)):
            raise TypeError("`ncomp` must be an int, float or a tuple in the "
                            "ADI case")

        if isinstance(ncomp, (int, float)):
            if isinstance(ncomp, int) and ncomp > n:
                ncomp = min(ncomp, n)
                print('Number of PCs too high (max PCs={}), using {} PCs '
                      'instead.'.format(n, ncomp))

            if source_xy is None:
                residuals_result = _project_subtract(cube, None, ncomp, scaling,
                                                     mask_center_px, svd_mode,
                                                     verbose, full_output,
                                                     cube_sig=cube_sig)
                if verbose:
                    timing(start_time)
                if full_output:
                    residuals_cube = residuals_result[0]
                    reconstructed = residuals_result[1]
                    V = residuals_result[2]
                    pcs = reshape_matrix(V, y, x)
                    recon = reshape_matrix(reconstructed, y, x)
                else:
                    residuals_cube = residuals_result

            # A rotation threshold is applied
            else:
                if delta_rot is None or fwhm is None:
                    msg = 'Delta_rot or fwhm parameters missing. Needed for the'
                    msg += 'PA-based rejection of frames from the library'
                    raise TypeError(msg)
                nfrslib = []
                residuals_cube = np.zeros_like(cube)
                recon_cube = np.zeros_like(cube)
                yc, xc = frame_center(cube[0], False)
                x1, y1 = source_xy
                ann_center = dist(yc, xc, y1, x1)
                pa_thr = _compute_pa_thresh(ann_center, fwhm, delta_rot)
                mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list))/2
                if pa_thr >= mid_range - mid_range * 0.1:
                    new_pa_th = float(mid_range - mid_range * 0.1)
                    if verbose:
                        msg = 'PA threshold {:.2f} is too big, will be set to '
                        msg += '{:.2f}'
                        print(msg.format(pa_thr, new_pa_th))
                    pa_thr = new_pa_th

                for frame in range(n):
                    if ann_center > fwhm * 3:  # TODO: 3 optimal value? new par?
                        ind = _find_indices_adi(angle_list, frame, pa_thr,
                                                truncate=True)
                    else:
                        ind = _find_indices_adi(angle_list, frame, pa_thr)

                    res_result = _project_subtract(cube, None, ncomp, scaling,
                                                   mask_center_px, svd_mode,
                                                   verbose, full_output, ind,
                                                   frame, cube_sig=cube_sig)
                    if full_output:
                        nfrslib.append(res_result[0])
                        residual_frame = res_result[1]
                        recon_frame = res_result[2]
                        residuals_cube[frame] = residual_frame.reshape((y, x))
                        recon_cube[frame] = recon_frame.reshape((y, x))
                    else:
                        nfrslib.append(res_result[0])
                        residual_frame = res_result[1]
                        residuals_cube[frame] = residual_frame.reshape((y, x))

                # number of frames in library printed for each annular quadrant
                if verbose:
                    descriptive_stats(nfrslib, verbose=verbose,
                                      label='Size LIB: ')

            residuals_cube_ = cube_derotate(residuals_cube, angle_list,
                                            imlib=imlib,
                                            interpolation=interpolation)
            frame = cube_collapse(residuals_cube_, mode=collapse, w=weights)
            if verbose:
                print('Done de-rotating and combining')
                timing(start_time)
            if source_xy is not None:
                return recon_cube, residuals_cube, residuals_cube_, frame
            else:
                return pcs, recon, residuals_cube, residuals_cube_, frame

        # When ncomp is a tuple, pca_grid is called
        else:
            gridre = pca_grid(cube, angle_list, fwhm, range_pcs=ncomp,
                              source_xy=source_xy, cube_ref=None, mode='fullfr',
                              svd_mode=svd_mode, scaling=scaling,
                              mask_center_px=mask_center_px, fmerit='mean',
                              imlib=imlib, interpolation=interpolation,
                              collapse=collapse, verbose=verbose,
                              full_output=full_output, debug=False,
                              plot=verbose, start_time=start_time, 
                              weights=weights)
            return gridre


def _adimsdi_singlepca(cube, angle_list, scale_list, ncomp, fwhm, source_xy,
                       scaling, mask_center_px, svd_mode, imlib, imlib2, 
                       interpolation, collapse, ifs_collapse_range, verbose, 
                       start_time, crop_ifs, batch, full_output, weights=None):
    """ Handles the full-frame ADI+mSDI single PCA post-processing.
    """
    z, n, y_in, x_in = cube.shape

    angle_list = check_pa_vector(angle_list)
    if not angle_list.shape[0] == n:
        msg = "Angle list vector has wrong length. It must equal the number"
        msg += " frames in the cube"
        raise ValueError(msg)

    if scale_list is None:
        raise ValueError('`scale_list` must be provided')
    else:
        check_array(scale_list, dim=1, msg='scale_list')
        if not scale_list.shape[0] == z:
            raise ValueError('`scale_list` has wrong length')

    scale_list = check_scal_vector(scale_list)
    big_cube = []

    if verbose:
        print('Rescaling the spectral channels to align the speckles')
    for i in Progressbar(range(n), verbose=verbose):
        cube_resc = scwave(cube[:, i, :, :], scale_list, imlib=imlib2,
                           interpolation=interpolation)[0]
        if crop_ifs:
            cube_resc = cube_crop_frames(cube_resc, size=y_in, verbose=False)
        big_cube.append(cube_resc)

    big_cube = np.array(big_cube)
    big_cube = big_cube.reshape(z * n, big_cube.shape[2], big_cube.shape[3])

    if verbose:
        timing(start_time)
        print('{} total frames'.format(n * z))
        print('Performing single-pass PCA')

    if isinstance(ncomp, (int, float)):
        # When ncomp is a int and batch is not None, incremental ADI-PCA is run
        if batch is not None:
            res_cube = pca_incremental(big_cube, angle_list, batch, ncomp, imlib,
                                       interpolation, collapse, verbose,
                                       return_residuals=True,
                                       start_time=start_time, weights=weights)
        # When ncomp is a int/float and batch is None, standard ADI-PCA is run
        else:
            res_cube = _project_subtract(big_cube, None, ncomp, scaling,
                                         mask_center_px, svd_mode, verbose,
                                         False)

        if verbose:
            timing(start_time)

        resadi_cube = np.zeros((n, y_in, x_in))

        if verbose:
            print('Descaling the spectral channels')
        if ifs_collapse_range == 'all':
            idx_ini = 0
            idx_fin = z
        else:
            idx_ini = ifs_collapse_range[0]
            idx_fin = ifs_collapse_range[1]
            
        for i in Progressbar(range(n), verbose=verbose):
            frame_i = scwave(res_cube[i*z+idx_ini:i*z+idx_fin, :, :], 
                             scale_list[idx_ini:idx_fin],
                             full_output=False, inverse=True, y_in=y_in,
                             x_in=x_in, imlib=imlib2,
                             interpolation=interpolation, collapse=collapse)
            resadi_cube[i] = frame_i

        if verbose:
            print('De-rotating and combining residuals')
            timing(start_time)
        der_res = cube_derotate(resadi_cube, angle_list, imlib=imlib,
                                interpolation=interpolation)
        frame = cube_collapse(der_res, mode=collapse, w=weights)
        cube_allfr_residuals = res_cube
        cube_adi_residuals = resadi_cube
        return cube_allfr_residuals, cube_adi_residuals, frame

    # When ncomp is a tuple, pca_grid is called
    elif isinstance(ncomp, tuple):
        gridre = pca_grid(big_cube, angle_list, fwhm, range_pcs=ncomp,
                          source_xy=source_xy, cube_ref=None, mode='fullfr',
                          svd_mode=svd_mode, scaling=scaling,
                          mask_center_px=mask_center_px, fmerit='mean',
                          imlib=imlib, interpolation=interpolation,
                          collapse=collapse, 
                          ifs_collapse_range=ifs_collapse_range, 
                          verbose=verbose, full_output=full_output, debug=False,
                          plot=verbose, start_time=start_time,
                          scale_list=scale_list, initial_4dshape=cube.shape,
                          weights=weights)
        return gridre

    else:
        raise TypeError("`ncomp` must be an int, float or a tuple for "
                        "single-pass PCA")


def _adimsdi_doublepca(cube, angle_list, scale_list, ncomp, scaling,
                       mask_center_px, svd_mode, imlib, imlib2, interpolation,
                       collapse, ifs_collapse_range, verbose, start_time, nproc,
                       weights=None, fwhm=4, conv=False, mask_rdi=None, 
                       cube_sig=None):
    """
    Handle the full-frame ADI+mSDI double PCA post-processing.

    """
    z, n, y_in, x_in = cube.shape

    global ARRAY
    ARRAY = cube  # to be passed to _adimsdi_doublepca_ifs

    if not isinstance(ncomp, tuple):
        raise TypeError("`ncomp` must be a tuple when a double pass PCA"
                        " is performed")
    else:
        ncomp_ifs, ncomp_adi = ncomp

    angle_list = check_pa_vector(angle_list)
    if not angle_list.shape[0] == n:
        msg = "Angle list vector has wrong length. It must equal the number"
        msg += " frames in the cube"
        raise ValueError(msg)

    if scale_list is None:
        raise ValueError('Scaling factors vector must be provided')
    else:
        if np.array(scale_list).ndim > 1:
            raise ValueError('Scaling factors vector is not 1d')
        if not scale_list.shape[0] == cube.shape[0]:
            raise ValueError('Scaling factors vector has wrong length')

    scale_list = check_scal_vector(scale_list)

    if verbose:
        print('{} spectral channels in IFS cube'.format(z))
        if ncomp_ifs is None:
            print('Combining multi-spectral frames (skipping PCA)')
        else:
            print('First PCA stage exploiting spectral variability')

    if ncomp_ifs is not None and ncomp_ifs > z:
        ncomp_ifs = min(ncomp_ifs, z)
        msg = 'Number of PCs too high (max PCs={}), using {} PCs instead'
        print(msg.format(z, ncomp_ifs))

    res = pool_map(nproc, _adimsdi_doublepca_ifs, iterable(range(n)), ncomp_ifs,
                   scale_list, scaling, mask_center_px, svd_mode, imlib2, 
                   interpolation, collapse, ifs_collapse_range, fwhm, conv,
                   mask_rdi)
    residuals_cube_channels = np.array(res)

    if verbose:
        timing(start_time)

    # de-rotation of the PCA processed channels, ADI fashion
    if ncomp_adi is None:
        if verbose:
            print('{} ADI frames'.format(n))
            print('De-rotating and combining frames (skipping PCA)')
        residuals_cube_channels_ = cube_derotate(residuals_cube_channels,
                                                 angle_list, imlib=imlib,
                                                 interpolation=interpolation)
        frame = cube_collapse(residuals_cube_channels_, mode=collapse, 
                              w=weights)
        if verbose:
            timing(start_time)
    else:
        if ncomp_adi > n:
            ncomp_adi = n
            print('Number of PCs too high, using  maximum of {} PCs '
                  'instead'.format(n))
        if verbose:
            print('{} ADI frames'.format(n))
            print('Second PCA stage exploiting rotational variability')

        res_ifs_adi = _project_subtract(residuals_cube_channels, None,
                                        ncomp_adi, scaling, mask_center_px,
                                        svd_mode, verbose=False,
                                        full_output=False, cube_sig=cube_sig)
        if verbose:
            print('De-rotating and combining residuals')
        der_res = cube_derotate(res_ifs_adi, angle_list, imlib=imlib,
                                interpolation=interpolation)
        residuals_cube_channels_ = der_res
        frame = cube_collapse(residuals_cube_channels_, mode=collapse, 
                              w=weights)
        if verbose:
            timing(start_time)
    return residuals_cube_channels, residuals_cube_channels_, frame


def _adimsdi_doublepca_ifs(fr, ncomp, scale_list, scaling, mask_center_px,
                           svd_mode, imlib, interpolation, collapse, 
                           ifs_collapse_range, fwhm, conv, mask_rdi=None):
    """
    Called by _adimsdi_doublepca with pool_map.
    """
    global ARRAY

    z, n, y_in, x_in = ARRAY.shape
    multispec_fr = ARRAY[:, fr, :, :]

    if ifs_collapse_range == 'all':
        idx_ini = 0
        idx_fin = z
    else:
        idx_ini = ifs_collapse_range[0]
        idx_fin = ifs_collapse_range[1]

    if ncomp is None:
        frame_i = cube_collapse(multispec_fr[idx_ini:idx_fin])
    else:
        cube_resc = scwave(multispec_fr, scale_list, imlib=imlib, 
                           interpolation=interpolation)[0]
        if conv:
            # convolve all frames with the same kernel
            cube_resc = cube_filter_lowpass(cube_resc, mode='gauss', 
                                            fwhm_size=fwhm, verbose=False)
        if mask_rdi is None:
            residuals = _project_subtract(cube_resc, None, ncomp, scaling,
                                          mask_center_px, svd_mode, 
                                          verbose=False, full_output=False)
        else:
            residuals = np.zeros_like(cube_resc)
            for i in range(z):
                cube_tmp = np.array([cube_resc[i]])
                cube_ref = np.array([cube_resc[j] for j in range(z) if j!=i])
                residuals[i] = cube_subtract_sky_pca(cube_tmp, cube_ref, 
                                                     mask_rdi, ncomp=ncomp, 
                                                     full_output=False)        

        frame_i = scwave(residuals[idx_ini:idx_fin], scale_list[idx_ini:idx_fin], 
                         full_output=False, inverse=True, y_in=y_in, x_in=x_in,
                         imlib=imlib, interpolation=interpolation, 
                         collapse=collapse)
        if mask_center_px:
            frame_i = mask_circle(frame_i, mask_center_px)

    return frame_i


def _adi_rdi_pca(cube, cube_ref, angle_list, ncomp, scaling, mask_center_px,
                 svd_mode, imlib, interpolation, collapse, verbose, start_time,
                 weights=None, mask_rdi=None, cube_sig=None):
    """ Handles the ADI+RDI post-processing.
    """
    n, y, x = cube.shape
    n_ref, y_ref, x_ref = cube_ref.shape
    angle_list = check_pa_vector(angle_list)
    if not isinstance(ncomp, int):
        raise TypeError("`ncomp` must be an int in the ADI+RDI case")
    if ncomp > n_ref:
        msg = 'Requested number of PCs ({}) higher than the number of frames '+\
              'in the reference cube ({}); using the latter instead.'
        print(msg.format(ncomp,n_ref))
        ncomp = n_ref

    if not cube_ref.ndim == 3:
        msg = 'Input reference array is not a cube or 3d array'
        raise ValueError(msg)
    if not y_ref == y and x_ref == x:
        msg = 'Reference and target frames have different shape'
        raise TypeError(msg)

    if mask_rdi is None:
        residuals_result = _project_subtract(cube, cube_ref, ncomp, scaling,
                                             mask_center_px, svd_mode, verbose,
                                             True, cube_sig=cube_sig)
        residuals_cube = residuals_result[0]
        reconstructed = residuals_result[1]
        V = residuals_result[2]
        pcs = reshape_matrix(V, y, x)
        recon = reshape_matrix(reconstructed, y, x)
    else:
        residuals_result = cube_subtract_sky_pca(cube, cube_ref, mask_rdi,
                                                 ncomp=ncomp, full_output=True)
        residuals_cube = residuals_result[0]
        pcs = residuals_result[2]
        recon = residuals_result[-1]
        
    residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                    interpolation=interpolation)
    frame = cube_collapse(residuals_cube_, mode=collapse, w=weights)
    if mask_center_px:
        frame = mask_circle(frame, mask_center_px)

    if verbose:
        print('Done de-rotating and combining')
        timing(start_time)

    return pcs, recon, residuals_cube, residuals_cube_, frame


def _project_subtract(cube, cube_ref, ncomp, scaling, mask_center_px,
                      svd_mode, verbose, full_output, indices=None, frame=None,
                      cube_sig=None):
    """
    PCA projection and model PSF subtraction. Used as a helping function by
    each of the PCA modes (ADI, ADI+RDI, ADI+mSDI).

    Parameters
    ----------
    cube : numpy ndarray
        Input cube.
    cube_ref : numpy ndarray
        Reference cube.
    ncomp : int
        Number of principal components.
    scaling : str
        Scaling of pixel values. See ``pca`` docstrings.
    mask_center_px : int
        Masking out a centered circular aperture.
    svd_mode : str
        Mode for SVD computation. See ``pca`` docstrings.
    verbose : bool
        Verbosity.
    full_output : bool
        Whether to return intermediate arrays or not.
    indices : list
        Indices to be used to discard frames (a rotation threshold is used).
    frame : int
        Index of the current frame (when indices is a list and a rotation
        threshold was applied).
    cube_sig: numpy ndarray, opt
        Cube with estimate of significant authentic signals. If provided, this
        will be subtracted from both the cube and the PCA library, before 
        projecting the cube onto the principal components.
        
    Returns
    -------
    ref_lib_shape : int
        [indices is not None, frame is not None] Number of
        rows in the reference library for the given frame.
    residuals: numpy ndarray
        Residuals, returned in every case.
    reconstructed : numpy ndarray
        [full_output=True] The reconstructed array.
    V : numpy ndarray
        [full_output=True, indices is None, frame is None] The right singular
        vectors of the input matrix, as returned by ``svd/svd_wrapper()``
    """
    _, y, x = cube.shape
    if isinstance(ncomp, int):
        # if cube_sig is not None:
        #     cube_emp = cube-cube_sig
        # else:
        #     cube_emp = None
        if indices is not None and frame is not None:
            matrix = prepare_matrix(cube, scaling, mask_center_px,
                                    mode='fullfr', verbose=False)
        else:
            matrix = prepare_matrix(cube, scaling, mask_center_px,
                                    mode='fullfr', verbose=verbose)
        if cube_sig is None:
            matrix_emp = matrix.copy()
        else:
            nfr = cube_sig.shape[0]
            cube_sig = np.reshape(cube_sig, (nfr, -1))
            matrix_emp = matrix-cube_sig

        if cube_ref is not None:
            ref_lib = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     mode='fullfr', verbose=verbose)
        else:
            ref_lib = matrix_emp

        # a rotation threshold is used (frames are processed one by one)
        if indices is not None and frame is not None:
            ref_lib = ref_lib[indices]
            if ref_lib.shape[0] <= 10:
                raise RuntimeError('Less than 10 frames left in the PCA library'
                                   ', Try decreasing the parameter delta_rot')
            curr_frame = matrix[frame]  # current frame
            curr_frame_emp = matrix_emp[frame]
            V = svd_wrapper(ref_lib, svd_mode, ncomp, False)
            transformed = np.dot(curr_frame_emp, V.T)
            reconstructed = np.dot(transformed.T, V)
            residuals = curr_frame - reconstructed
            if full_output:
                return ref_lib.shape[0], residuals, reconstructed
            else:
                return ref_lib.shape[0], residuals

        # the whole matrix is processed at once
        else:
            V = svd_wrapper(ref_lib, svd_mode, ncomp, verbose)
            transformed = np.dot(V, matrix_emp.T)
            reconstructed = np.dot(transformed.T, V)
            residuals = matrix - reconstructed
            residuals_res = reshape_matrix(residuals, y, x)
            if full_output:
                return residuals_res, reconstructed, V
            else:
                return residuals_res

    elif isinstance(ncomp, float):
        if not 1 > ncomp > 0:
            raise ValueError("when `ncomp` if float, it must lie in the "
                             "interval (0,1]")

        svdecomp = SVDecomposer(cube, mode='fullfr', svd_mode=svd_mode,
                                scaling=scaling, verbose=verbose)
        _ = svdecomp.get_cevr(plot=False)
        # in this case ncomp is the desired CEVR
        cevr = ncomp
        ncomp = svdecomp.cevr_to_ncomp(cevr)
        V = svdecomp.v[:ncomp]
        transformed = np.dot(V, svdecomp.matrix.T)
        reconstructed = np.dot(transformed.T, V)
        residuals = svdecomp.matrix - reconstructed
        residuals_res = reshape_matrix(residuals, y, x)

        if verbose and isinstance(cevr, float):
            print("Components used : {}".format(V.shape[0]))

        if full_output:
            return residuals_res, reconstructed, V
        else:
            return residuals_res


#! /usr/bin/env python

"""
Module with various functions to create shapes, annuli and segments.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['dist',
           'dist_matrix',
           'frame_center',
           'get_square',
           'get_circle',
           'get_ellipse',
           'get_annulus_segments',
           'get_annular_wedge',
           'get_ell_annulus',
           'mask_circle',
           'create_ringed_spider_mask',
           'matrix_scaling',
           'prepare_matrix',
           'reshape_matrix',
           'cart_to_pol',
           'pol_to_cart']

import numpy as np
from skimage.draw import polygon
from skimage.draw import disk
from sklearn.preprocessing import scale

from ..conf.utils_conf import frame_or_shape


def mask_circle(array, radius, fillwith=0, mode='in'):
    """
    Mask the pixels inside/outside of a centered circle with ``fillwith``.

    Returns a modified copy of ``array``.

    Parameters
    ----------
    array : 2d/3d/4d numpy ndarray
        Input frame or cube.
    radius : int
        Radius of the circular aperture.
    fillwith : int, float or np.nan, optional
        Value to put instead of the masked out pixels.
    mode : {'in', 'out'}, optional
        When set to 'in' then the pixels inside the radius are set to
        ``fillwith``. When set to 'out' the pixels outside the circular mask are
        set to ``fillwith``.

    Returns
    -------
    array_masked : numpy ndarray
        Masked frame or cube.

    """
    if not isinstance(fillwith, (int, float)):
        raise ValueError('`fillwith` must be integer, float or np.nan')

    cy, cx = frame_center(array)

    shape = (array.shape[-2],array.shape[-1])
    ind = disk((cy, cx), radius, shape=shape)

    if mode == 'in':
        array_masked = array.copy()
        if array.ndim == 2:
            array_masked[ind] = fillwith
        elif array.ndim == 3:
            array_masked[:, ind[1], ind[0]] = fillwith
        elif array.ndim == 4:
            array_masked[:, :, ind[1], ind[0]] = fillwith

    elif mode == 'out':
        array_masked = np.full_like(array, fillwith)
        if array.ndim == 2:
            array_masked[ind] = array[ind]
        elif array.ndim == 3:
            array_masked[:, ind[1], ind[0]] = array[:, ind[1], ind[0]]
        elif array.ndim == 4:
            array_masked[:, :, ind[1], ind[0]] = array[:, :, ind[1], ind[0]]

    return array_masked


def create_ringed_spider_mask(im_shape, ann_out, ann_in=0, sp_width=10,
                              sp_angle=0, nlegs=6):
    """
    Mask out information outside the annulus and inside the spiders (zeros).

    Parameters
    ----------
    im_shape : tuple of int
        Tuple of length two with 2d array shape (Y,X).
    ann_out : int
        Outer radius of the annulus.
    ann_in : int, opt
        Inner radius of the annulus.
    sp_width : int, opt
        Width of the spider arms (6 legs by default).
    sp_angle : int, opt
        angle of the first spider arm (on the positive horizontal axis) in
        counter-clockwise sense.
    nlegs: int, opt
        Number of legs of the spider.

    Returns
    -------
    mask : numpy ndarray
        2d array of zeros and ones.

    """
    mask = np.zeros(im_shape)
    nbranch = int(nlegs/2)

    s = im_shape
    r = min(s)/2
    theta = np.arctan2(sp_width/2, r)

    cy, cx = frame_center(mask)
    rr0, cc0 = disk((cy, cx), ann_out)
    # check that all indices are within frame dimensions
    cond1 = rr0>=0
    cond2 = rr0<s[0]
    cond3 = cc0>=0
    cond4 = cc0<s[1]
    all_cond = cond1 & cond2 & cond3 & cond4
    rr0 = rr0[np.where(all_cond)]
    cc0 = cc0[np.where(all_cond)]
    mask[rr0, cc0] = 1

    t0 = np.array([theta, np.pi-theta, np.pi+theta, np.pi*2 - theta])
    if isinstance(sp_angle, (list,np.ndarray)):
        dtheta = [sp_angle[i]-sp_angle[0] for i in range(nbranch)]
    else:
        sp_angle = [sp_angle]
        dtheta = [i*180./nbranch for i in range(nbranch)]
    tn = np.zeros([nbranch,4])
    xn = np.zeros_like(tn)
    yn = np.zeros_like(tn)
    for i in range(nbranch):
        tn[i] = t0 + np.deg2rad(sp_angle[0] + dtheta[i])
        xn[i] = r * np.cos(tn[i]) + s[1]/2
        yn[i] = r * np.sin(tn[i]) + s[0]/2
        rrn, ccn = polygon(yn[i], xn[i])
        mask[rrn, ccn] = 0
    rr4, cc4 = disk((cy, cx), ann_in)
    mask[rr4, cc4] = 0
    
    return mask


def dist(yc, xc, y1, x1):
    """
    Return the Euclidean distance between two points, or between an array 
    of positions and a point.
    """
    return np.sqrt(np.power(yc-y1,2) + np.power(xc-x1,2))


def dist_matrix(n, cx=None, cy=None):
    """
    Create matrix with euclidian distances from a reference point (cx, cy).

    Parameters
    ----------
    n : int
        output image shape is (n, n)
    cx,cy : float
        reference point. Defaults to the center.

    Returns
    -------
    im : ndarray with shape (n, n)

    Notes
    -----
    This is a replacement for ANDROMEDA's DISTC.

    """
    if cx is None:
        cx = (n - 1) / 2
    if cy is None:
        cy = (n - 1) / 2

    yy, xx = np.ogrid[:n, :n]
    return np.sqrt((yy-cy)**2 + (xx-cx)**2)


def frame_center(array, verbose=False):
    """
    Return the coordinates y,x of the frame(s) center.

    Parameters
    ----------
    array : 2d/3d/4d numpy ndarray
        Frame or cube.
    verbose : bool optional
        If True the center coordinates are printed out.

    Returns
    -------
    cy, cx : float
        Coordinates of the center.

    """
    if array.ndim == 2:
        shape = array.shape
    elif array.ndim == 3:
        shape = array[0].shape
    elif array.ndim == 4:
        shape = array[0, 0].shape
    else:
        raise ValueError('`array` is not a 2d, 3d or 4d array')

    cy = shape[0] / 2 - 0.5
    cx = shape[1] / 2 - 0.5

    if verbose:
        print('Center px coordinates at x,y = ({}, {})'.format(cx, cy))  
    
    return cy, cx


def get_square(array, size, y, x, position=False, force=False, verbose=True):
    """
    Return an square subframe from a 2d array or image.

    Parameters
    ----------
    array : 2d numpy ndarray
        Input frame.
    size : int
        Size of the subframe.
    y : int
        Y coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    x : int
        X coordinate of the center of the subframe (obtained with the function
        ``frame_center``).
    position : bool, optional
        If set to True return also the coordinates of the bottom-left vertex.
    force : bool, optional
        Size and the size of the 2d array must be both even or odd. With
        ``force`` set to True this condition can be avoided.
    verbose : bool optional
        If True, warning messages might be shown.

    Returns
    -------
    array_out : numpy ndarray
        Sub array.
    y0, x0 : int
        [position=True] Coordinates of the bottom-left vertex.

    """
    size_init_y = array.shape[0]
    size_init_x = array.shape[1]
    size_init = array.shape[0] # "force" cases assume square input frame
    
    if array.ndim != 2:
        raise TypeError('Input array is not a 2d array.')
    if not isinstance(size, int):
        raise TypeError('`Size` must be integer')
    if size >= size_init_y and size >= size_init_x:  # assuming square frames
        msg = "`Size` is equal to or bigger than the initial frame size"
        raise ValueError(msg)

    if not force:
        # Even input size
        if size_init % 2 == 0:
            # Odd size
            if size % 2 != 0:
                size += 1
                if verbose:
                    print("`Size` is odd (while input frame size is even). "
                          "Setting `size` to {} pixels".format(size))
        # Odd input size
        else:
            # Even size
            if size % 2 == 0:
                size += 1
                if verbose:
                    print("`Size` is even (while input frame size is odd). "
                          "Setting `size` to {} pixels".format(size))
    else:
        # Even input size
        if size_init % 2 == 0:
            # Odd size
            if size % 2 != 0 and verbose:
                print("WARNING: `size` is odd while input frame size is even. "
                      "Make sure the center coordinates are set properly")
        # Odd input size
        else:
            # Even size
            if size % 2 == 0 and verbose:
                print("WARNING: `size` is even while input frame size is odd. "
                      "Make sure the center coordinates are set properly")

    # wing is added to the sides of the subframe center
    wing = (size - 1) / 2

    y0 = int(y - wing)
    y1 = int(y + wing + 1)  # +1 cause endpoint is excluded when slicing
    x0 = int(x - wing)
    x1 = int(x + wing + 1)

    if y0 < 0 or x0 < 0 or y1 > size_init_y or x1 > size_init_x:
        # assuming square frames
        raise RuntimeError('square cannot be obtained with size={}, y={}, x={}'
                           ''.format(size, y, x))

    array_out = array[y0: y1, x0: x1].copy()

    if position:
        return array_out, y0, x0
    else:
        return array_out


def get_circle(array, radius, cy=None, cx=None, mode="mask"):
    """
    Return a centered circular region from a 2d ndarray.

    Parameters
    ----------
    array : numpy ndarray
        Input 2d array or image.
    radius : int
        The radius of the circular region.
    cy, cx : int, optional
        Coordinates of the circle center. If one of them is ``None``, the center
        of ``array`` is used.
    mode : {'mask', 'val'}, optional
        Controls what is returned: array with circular mask applied, or values
        of the pixels in the circular region.

    Returns
    -------
    masked : numpy ndarray
        [mode="mask"] Input array with the circular mask applied.
    values : numpy ndarray
        [mode="val"] 1d array with the values of the pixels in the circular
        region.

    Notes
    -----
    An alternative implementation would use ``skimage.draw.circle``. ``circle``
    performs better on large ``array``s (e.g. 1000px, 10.000px), while the
    current implementation is faster for small ``array``s (e.g. 100px). See
    `test_shapes.py` for benchmark details.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')
    sy, sx = array.shape
    if cy is None or cx is None:
        cy, cx = frame_center(array, verbose=False)

    # ogrid is a multidim mesh creator (faster than mgrid):
    yy, xx = np.ogrid[:sy, :sx]
    circle = (yy - cy) ** 2 + (xx - cx) ** 2  # eq of circle. sq dist to center
    circle_mask = circle < radius ** 2  # boolean mask
    if mode == "mask":
        return array * circle_mask
    elif mode == "val":
        return array[circle_mask]
    else:
        raise ValueError("mode '{}' unknown!".format(mode))


def get_ellipse(data, a, b, pa, cy=None, cx=None, mode="ind"):
    """
    Return a centered elliptical region from a 2d ndarray.

    Parameters
    ----------
    data : numpy ndarray or tuple
        Input 2d array (image) or tuple with a shape.
    a : float
        Semi-major axis.
    b : float
        Semi-minor axis.
    pa : deg, float
        The PA of the semi-major axis in degrees.
    cy, cx : int or None, optional
        Coordinates of the circle center. If ``None``, the center is determined
        by the ``frame_center`` function.
    mode : {'ind', 'val', 'mask', 'bool'}, optional
        Controls what is returned: indices of selected pixels, values of
        selected pixels, or a boolean mask.

    Returns
    -------
    indices : tuple(y, x)
        [mode='ind'] Coordinates of the inner elliptical region.
    values : 1d ndarray
        [mode='val'] Values of the pixels in the inner elliptical region.
    masked : 2d ndarray
        [mode='mask'] Input image where the outer region is masked with ``0``.
    bool_mask : 2d boolean ndarray
        [mode='bool'] A boolean mask where ``True`` is the inner region.

    """

    def distance(yc, xc, y1, x1):
        return np.sqrt((yc - y1) ** 2 + (xc - x1) ** 2)
    # --------------------------------------------------------------------------

    array = frame_or_shape(data)

    if cy is None or cx is None:
        cy, cx = frame_center(array, verbose=False)

    # Definition of other parameters of the ellipse
    f = np.sqrt(a ** 2 - b ** 2)  # dist between center and foci of the ellipse
    pa_rad = np.deg2rad(pa)
    pos_f1 = (cy + f * np.cos(pa_rad), cx + f * np.sin(pa_rad))  # first focus
    pos_f2 = (cy - f * np.cos(pa_rad), cx - f * np.sin(pa_rad))  # second focus

    # ogrid is a multidim mesh creator (faster than mgrid):
    yy, xx = np.ogrid[:array.shape[0], :array.shape[1]]
    ellipse = (distance(yy, xx, pos_f1[0], pos_f1[1]) +
               distance(yy, xx, pos_f2[0], pos_f2[1]))
    ellipse_mask = ellipse < 2 * a  # boolean mask

    if mode == "ind":
        return np.where(ellipse_mask)
    elif mode == "val":
        return array[ellipse_mask]
    elif mode == "mask":
        return array * ellipse_mask
    elif mode == "bool":
        return ellipse_mask
    else:
        raise ValueError("mode '{}' unknown!".format(mode))


def get_annulus_segments(data, inner_radius, width, nsegm=1, theta_init=0,
                         optim_scale_fact=1, mode="ind"):
    """
    Return indices or values in segments of a centerered annulus.

    The annulus is defined by ``inner_radius <= annulus < inner_radius+width``.

    Parameters
    ----------
    data : 2d numpy ndarray or tuple
        Input 2d array (image) ot tuple with its shape.
    inner_radius : float
        The inner radius of the donut region.
    width : float
        The size of the annulus.
    nsegm : int
        Number of segments of annulus to be extracted.
    theta_init : int
        Initial azimuth [degrees] of the first segment, counting from the
        positive x-axis counterclockwise.
    optim_scale_fact : float
        To enlarge the width of the segments, which can then be used as
        optimization segments (e.g. in LOCI).
    mode : {'ind', 'val', 'mask'}, optional
        Controls what is returned: indices of selected pixels, values of
        selected pixels, or a boolean mask.

    Returns
    -------
    indices : list of ndarrays
        [mode='ind'] Coordinates of pixels for each annulus segment.
    values : list of ndarrays
        [mode='val'] Pixel values.
    masked : list of ndarrays
        [mode='mask'] Copy of ``data`` with masked out regions.

    Notes
    -----
    Moving from ``get_annulus`` to ``get_annulus_segments``:

    .. code::python
        # get_annulus handles one single segment only, so note the ``[0]`` after
        the call to get_annulus_segments if you want to work with one single
        segment only.

        get_annulus(arr, 2, 3, output_indices=True)
        # is the same as
        get_annulus_segments(arr, 2, 3)[0]

        get_annulus(arr, inr, w, output_values=True)
        # is the same as
        get_annulus_segments(arr, inr, w, mode="val")[0]

        get_annulus(arr, inr, w)
        # is the same as
        get_annulus_segments(arr, inr, w, mode="mask")[0]

        # the only difference is the handling of the border values:
        # get_annulus_segments is `in <= ann < out`, while get_annulus is
        # `in <= ann <= out`. But that should make no difference in practice.

    """
    array = frame_or_shape(data)

    if not isinstance(nsegm, int):
        raise TypeError('`nsegm` must be an integer')

    cy, cx = frame_center(array)
    azimuth_coverage = np.deg2rad(int(np.ceil(360 / nsegm)))
    twopi = 2 * np.pi

    yy, xx = np.mgrid[:array.shape[0], :array.shape[1]]
    rad = np.sqrt((xx - cx) ** 2 + (yy - cy) ** 2)
    phi = np.arctan2(yy - cy, xx - cx)
    phirot = phi % twopi
    outer_radius = inner_radius + (width*optim_scale_fact)
    masks = []

    for i in range(nsegm):
        phi_start = np.deg2rad(theta_init) + (i * azimuth_coverage)
        phi_end = phi_start + azimuth_coverage

        if phi_start < twopi and phi_end > twopi:
            masks.append((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start) & (phirot <= twopi) |
                         (rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= 0) & (phirot < phi_end - twopi))
        elif phi_start >= twopi and phi_end > twopi:
            masks.append((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start - twopi) &
                         (phirot < phi_end - twopi))
        else:
            masks.append((rad >= inner_radius) & (rad < outer_radius) &
                         (phirot >= phi_start) & (phirot < phi_end))

    if mode == "ind":
        return [np.where(mask) for mask in masks]
    elif mode == "val":
        return [array[mask] for mask in masks]
    elif mode == "mask":
        return [array*mask for mask in masks]
    else:
        raise ValueError("mode '{}' unknown!".format(mode))


def get_annular_wedge(data, inner_radius, width, wedge=(0,360), mode="ind"):
    """
    Return indices or values in segments of a centered annulus.

    The annulus is defined by ``inner_radius <= annulus < inner_radius+width``.

    Parameters
    ----------
    data : 2d numpy ndarray or tuple
        Input 2d array (image) ot tuple with its shape.
    inner_radius : float
        The inner radius of the donut region.
    width : float
        The size of the annulus.
    wedge : tuple of 2 floats
        Initial and final azimuths [degrees] of the annular segment, counting
        from the positive x-axis counter-clockwise.
    mode : {'ind', 'val', 'mask'}, optional
        Controls what is returned: indices of selected pixels, values of
        selected pixels, or a boolean mask.

    Returns
    -------
    indices : list of ndarrays
        [mode='ind'] Coordinates of pixels for each annulus segment.
    values : list of ndarrays
        [mode='val'] Pixel values.
    masked : list of ndarrays
        [mode='mask'] Copy of ``data`` with masked out regions.

    Notes
    -----
    Moving from ``get_annulus`` to ``get_annulus_segments``:

    .. code::python
        # get_annulus handles one single segment only, so note the ``[0]`` after
        the call to get_annulus_segments if you want to work with one single
        segment only.

        get_annulus(arr, 2, 3, output_indices=True)
        # is the same as
        get_annulus_segments(arr, 2, 3)[0]

        get_annulus(arr, inr, w, output_values=True)
        # is the same as
        get_annulus_segments(arr, inr, w, mode="val")[0]

        get_annulus(arr, inr, w)
        # is the same as
        get_annulus_segments(arr, inr, w, mode="mask")[0]

        # the only difference is the handling of the border values:
        # get_annulus_segments is `in <= ann < out`, while get_annulus is
        # `in <= ann <= out`. But that should make no difference in practice.

    """
    array = frame_or_shape(data)

    cy, cx = frame_center(array)
    azimuth_coverage = np.deg2rad(wedge[1]-wedge[0])
    twopi = 2 * np.pi

    yy, xx = np.mgrid[:array.shape[0], :array.shape[1]]
    rad = np.sqrt((xx - cx) ** 2 + (yy - cy) ** 2)
    phi = np.arctan2(yy - cy, xx - cx)
    phirot = phi % twopi
    outer_radius = inner_radius + width

    phi_start = np.deg2rad(wedge[0])
    phi_end = phi_start + azimuth_coverage

    if phi_start < twopi and phi_end > twopi:
        mask = ((rad >= inner_radius) & (rad < outer_radius) &
                     (phirot >= phi_start) & (phirot <= twopi) |
                     (rad >= inner_radius) & (rad < outer_radius) &
                     (phirot >= 0) & (phirot < phi_end - twopi))
    elif phi_start >= twopi and phi_end > twopi:
        mask = ((rad >= inner_radius) & (rad < outer_radius) &
                     (phirot >= phi_start - twopi) &
                     (phirot < phi_end - twopi))
    else:
        mask = ((rad >= inner_radius) & (rad < outer_radius) &
                     (phirot >= phi_start) & (phirot < phi_end))

    if mode == "ind":
        return np.where(mask)
    elif mode == "val":
        return array[mask]
    elif mode == "mask":
        return array*mask
    else:
        raise ValueError("mode '{}' unknown!".format(mode))
        

def get_ell_annulus(data, a, b, PA, width, cy=None, cx=None, mode="ind"):
    """
    Return a centered elliptical annulus from a 2d ndarray

    All the rest pixels are set to zeros.

    Parameters
    ----------
    data : numpy ndarray or tuple
        Input 2d array (image) or tuple with a shape.
    a : float
        Semi-major axis.
    b : float
        Semi-minor axis.
    PA : deg, float
        The PA of the semi-major axis in degrees.
    width : float
        The size of the annulus along the semi-major axis; it is proportionnally
        thinner along the semi-minor axis.
    output_values : {False, True}, optional
        If True returns the values of the pixels in the annulus.
    cy, cx : int or None, optional
        Coordinates of the circle center. If ``None``, the center is determined
        by the ``frame_center`` function.
    mode : {'ind', 'val', 'mask'}, optional
        Controls what is returned: indices of selected pixels, values of
        selected pixels, or a boolean mask.

    Returns
    -------
    indices : tuple(y, x)
        [mode='ind'] Coordinates of the inner elliptical region.
    values : 1d ndarray
        [mode='val'] Values of the pixels in the inner elliptical region.
    masked : 2d ndarray
        [mode='mask'] Input image where the outer region is masked with ``0``.

    """
    array = frame_or_shape(data)

    hwa = width / 2  # half width for a
    hwb = (width * b / a) / 2  # half width for b

    big_ellipse = get_ellipse(array, a + hwa, b + hwb, PA, cy=cy, cx=cx,
                              mode="bool")
    small_ellipse = get_ellipse(array, a - hwa, b - hwb, PA, cy=cy, cx=cx,
                                mode="bool")

    ell_ann_mask = big_ellipse ^ small_ellipse

    if mode == "ind":
        return np.where(ell_ann_mask)
    elif mode == "val":
        return array[ell_ann_mask]
    elif mode == "mask":
        return array * ell_ann_mask
    elif mode == "bool":
        return ell_ann_mask
    else:
        raise ValueError("mode '{}' unknown!".format(mode))


def matrix_scaling(matrix, scaling):
    """
    Scale a matrix using ``sklearn.preprocessing.scale`` function.

    Parameters
    ----------
    matrix : 2d numpy ndarray
        Input 2d array.
    scaling : None or string
        Scaling method.

        ``None``
            no scaling is performed on the input data before SVD
        ``"temp-mean"``
            temporal px-wise mean subtraction
        ``"spat-mean"``
            the spatial mean is subtracted
        ``temp-standard"``
            temporal mean centering plus scaling to unit variance
        ``"spat-standard"``
            spatial mean centering plus scaling to unit variance

    Returns
    -------
    matrix : 2d numpy ndarray
        2d array with scaled values.

    """
    if scaling is None:
        pass
    elif scaling == 'temp-mean':
        matrix = scale(matrix, with_mean=True, with_std=False)
    elif scaling == 'spat-mean':
        matrix = scale(matrix, with_mean=True, with_std=False, axis=1)
    elif scaling == 'temp-standard':
        matrix = scale(matrix, with_mean=True, with_std=True)
    elif scaling == 'spat-standard':
        matrix = scale(matrix, with_mean=True, with_std=True, axis=1)
    else:
        raise ValueError('Scaling mode not recognized')

    return matrix


def prepare_matrix(array, scaling=None, mask_center_px=None, mode='fullfr',
                   inner_radius=None, outer_radius=None, verbose=True):
    """
    Build the matrix for the SVD/PCA and other matrix decompositions.

    Center the data and mask the frame's central area if needed.

    Parameters
    ----------
    array : 3d numpy ndarray
        Input cube.
    scaling : {None, "temp-mean", spat-mean", "temp-standard", "spat-standard"},
        None or str optional
        Pixel-wise scaling mode using ``sklearn.preprocessing.scale`` function.
        If set to None, the input matrix is left untouched. Otherwise:

        ``temp-mean``: temporal px-wise mean is subtracted.

        ``spat-mean``: spatial mean is subtracted.

        ``temp-standard``: temporal mean centering plus scaling pixel values
        to unit variance.

        ``spat-standard``: spatial mean centering plus scaling pixel values
        to unit variance.

    mask_center_px : None or int, optional
        [mode='fullfr'] Whether to mask the center of the frames or not.
    mode : {'fullfr', 'annular'}, optional
        Whether to use the whole frames or a single annulus.
    inner_radius : int or float, optional
        [mode='annular'] Distance in pixels from the center of the frame to the
        inner radius of the annulus.
    outer_radius : int or float, optional
        [mode='annular'] Distance in pixels from the center of the frame to the
        outer radius of the annulus.
    verbose : bool, optional
        If True prints intermediate info.

    Returns
    -------
    matrix : 2d numpy ndarray
        Out matrix whose rows are vectorized frames from the input cube.
    ind : tuple
        [mode='annular'] Indices of the annulus as ``(yy, xx)``.

    """
    if mode == 'annular':
        if inner_radius is None or outer_radius is None:
            raise ValueError('`inner_radius` and `outer_radius` must be defined'
                             ' in annular mode')
        fr_size = array.shape[1]
        annulus_width = int(np.round(outer_radius - inner_radius))
        ind = get_annulus_segments((fr_size, fr_size), inner_radius,
                                   annulus_width, nsegm=1)[0]
        yy, xx = ind
        matrix = array[:, yy, xx]

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the cube annulus. Matrix shape: ({}, {})'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix, ind

    elif mode == 'fullfr':
        if mask_center_px:
            array = mask_circle(array, mask_center_px)

        nfr = array.shape[0]
        matrix = np.reshape(array, (nfr, -1))  # == for i: array[i].flatten()

        matrix = matrix_scaling(matrix, scaling)

        if verbose:
            msg = 'Done vectorizing the frames. Matrix shape: ({}, {})'
            print(msg.format(matrix.shape[0], matrix.shape[1]))
        return matrix


def reshape_matrix(array, y, x):
    """
    Convert a matrix whose rows are vect. frames to a cube with reshaped frames.

    Parameters
    ----------
    array : 2d ndarray
        Input data of shape ``(nframes, npixels)``. Every row (``array[n]``)
        corresponds to one vectorized ("flattened") 2d frame.
    y, x : int
        desired height and width of the frames. ``y*x = npixels``

    Returns
    -------
    cube : 3d ndarray
        Cube of shape ``(nframes, y, x)``.

    Examples
    --------
    .. code:: python

        In [1]: vect_frames = np.array([[1, 1, 1, 2, 2, 2], [1, 2, 3, 4, 5, 6]])
        In [2]: cube = vip.var.reshape_matrix(vect_frames, 2, 3)
        In [3]: cube
        Out[3]:
        array([[[1, 1, 1],
                [2, 2, 2]],

               [[1, 2, 3],
                [4, 5, 6]]])
        In [4]: cube.shape
        Out[4]: (2, 2, 3)

    """
    return array.reshape(array.shape[0], y, x)


def cart_to_pol(x, y, cx=0, cy=0):
    """
    Returns polar coordinates for input cartesian coordinates
    
    Parameters
    ----------
    x : float or numpy ndarray
        x coordinates with respect to the center
    y : float or numpy ndarray
        y coordinates with respect to the center

    Returns
    -------
    r, theta: floats or numpy ndarrays
        radii and polar angles (trigonometric) corresponding to the input
        x and y.
    """
    
    r = dist(cy,cx,y,x)
    theta = np.rad2deg(np.arctan2(y-cy,x-cx))
    
    return r, theta


def pol_to_cart(r, theta, cx=0, cy=0):
    """
    Returns cartesian coordinates for input polar coordinates
    
    Parameters
    ----------
    r, theta : float or numpy ndarray
        radii and polar angles (trigonometric) corresponding to the input
        x and y.
    cx, cy : float or numpy ndarray
        x, y coordinates of the center of the image to be considered for 
        conversion to cartesian coordinates.

    Returns
    -------
    x, y: floats or numpy ndarrays
        x, y coordinates corresponding to input radii and polar (trigonotetric)
        angles.
    """
    
    x = cx+r*np.cos(np.deg2rad(theta))
    y = cy+r*np.sin(np.deg2rad(theta))
    
    return x, y